"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InXE v8'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 8

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return []

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "ConfigNumber": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/ConfigNumber", 0, "ALL"
            ),
            "OutHtRCur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/OutHtRCur", 27, "ALL"
            ),
            "DirectCur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/DirectCur", 26, "ALL"
            ),
            "PumpTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/PumpTimeOut", 31, "ALL"
            ),
            "LightTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/LightTimeOut", 32, "ALL"
            ),
            "FiltInterface": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltInterface",
                17,
                None,
                ["PurgeOnly", "FiltCP", "FiltP1"],
                None,
                None,
                "ALL",
            ),
            "OTActSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/OTActSetpointG",
                33,
                "ALL",
            ),
            "OTTriggerG": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/AdvanceFilterConfig/OTTriggerG", 35, "ALL"
            ),
            "CPOTMaxOnTime": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/CPOTMaxOnTime",
                36,
                "ALL",
            ),
            "CPOTMaxOffTime": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/CPOTMaxOffTime",
                38,
                "ALL",
            ),
            "FiltOTDuration24H": GeckoWordStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOTDuration24H",
                40,
                "ALL",
            ),
            "FiltSuspendTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltSuspendTime",
                42,
                "ALL",
            ),
            "O3Pump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Pump",
                14,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "O3Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Type",
                15,
                None,
                ["Standard", "Toggle"],
                None,
                None,
                "ALL",
            ),
            "O3SuspendTime": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3SuspendTime", 43, "ALL"
            ),
            "SetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/SetpointG",
                1,
                "ALL",
            ),
            "HeaterPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeaterPump",
                16,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "TempUnits": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/TempUnits",
                18,
                None,
                ["F", "C"],
                None,
                None,
                "ALL",
            ),
            "MinSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/MinSetpointG",
                46,
                "ALL",
            ),
            "MaxSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/MaxSetpointG",
                48,
                "ALL",
            ),
            "EconType": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/EconType",
                50,
                None,
                ["Standard", "Night"],
                None,
                None,
                "ALL",
            ),
            "NbPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/NbPhases", 29, "ALL"
            ),
            "InputCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/InputCurrent", 30, "ALL"
            ),
            "FiltFreq": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltFreq", 4, "ALL"
            ),
            "FiltStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltStart",
                5,
                "ALL",
            ),
            "FiltDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltDur", 3, "ALL"
            ),
            "EconStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStart",
                7,
                "ALL",
            ),
            "EconDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/EconDur", 8, "ALL"
            ),
            "EconProgAvailable": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconProgAvailable",
                51,
                None,
                ["NA", "Available"],
                None,
                None,
                "ALL",
            ),
            "SoakOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SoakOnCustomKey",
                52,
                0,
                "ALL",
            ),
            "OffOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/OffOnCustomKey",
                52,
                1,
                "ALL",
            ),
            "EconControlableManually": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconControlableManually",
                52,
                2,
                "ALL",
            ),
            "TimeFormat": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/TimeFormat",
                19,
                None,
                ["NA", "AmPm", "24h"],
                None,
                None,
                "ALL",
            ),
            "AmbiantOHTrigADC": GeckoWordStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/AmbiantOHTrigADC", 44, "ALL"
            ),
        }



"""Gecko driver communication handlers."""

from .configfile import GeckoConfigFileProtocolHandler
from .firmware import GeckoUpdateFirmwareProtocolHandler
from .getchannel import GeckoGetChannelProtocolHandler
from .hello import GeckoHelloProtocolHandler
from .packcommand import GeckoPackCommandProtocolHandler
from .packet import GeckoPacketProtocolHandler
from .ping import GeckoPingProtocolHandler
from .reminders import GeckoRemindersProtocolHandler, GeckoReminderType
from .rferr import GeckoRFErrProtocolHandler
from .statusblock import (
    GeckoAsyncPartialStatusBlockProtocolHandler,
    GeckoStatusBlockProtocolHandler,
)
from .unhandled import GeckoUnhandledProtocolHandler
from .version import GeckoVersionProtocolHandler
from .watercare import (
    GeckoGetWatercareModeProtocolHandler,
    GeckoGetWatercareScheduleListProtocolHandler,
    GeckoSetWatercareModeProtocolHandler,
    GeckoWatercareErrorHandler,
    GeckoWatercareScheduleManager,
)

__all__ = [
    "GeckoAsyncPartialStatusBlockProtocolHandler",
    "GeckoConfigFileProtocolHandler",
    "GeckoGetChannelProtocolHandler",
    "GeckoGetWatercareModeProtocolHandler",
    "GeckoGetWatercareModeProtocolHandler",
    "GeckoGetWatercareScheduleListProtocolHandler",
    "GeckoHelloProtocolHandler",
    "GeckoPackCommandProtocolHandler",
    "GeckoPacketProtocolHandler",
    "GeckoPingProtocolHandler",
    "GeckoRFErrProtocolHandler",
    "GeckoReminderType",
    "GeckoRemindersProtocolHandler",
    "GeckoSetWatercareModeProtocolHandler",
    "GeckoStatusBlockProtocolHandler",
    "GeckoUnhandledProtocolHandler",
    "GeckoUpdateFirmwareProtocolHandler",
    "GeckoVersionProtocolHandler",
    "GeckoWatercareErrorHandler",
    "GeckoWatercareScheduleManager",
]



"""Gecko GETWC/WCGET/SETWC/WCSET/REQWC/WCREQ handlers."""

from __future__ import annotations

import logging
import struct
from typing import Any

from geckolib.config import GeckoConfig
from geckolib.const import GeckoConstants

from .packet import GeckoPacketProtocolHandler

GETWC_VERB = b"GETWC"  # Get watercare mode
WCGET_VERB = b"WCGET"  # Get watercare mode response
SETWC_VERB = b"SETWC"  # Set watercare mode
WCSET_VERB = b"WCSET"  # Set watercare mode response
REQWC_VERB = b"REQWC"  # Get watercare schedule list
WCREQ_VERB = b"WCREQ"  # Get watercare schedule list response
WCERR_VERB = b"WCERR"  # Watercare error


GET_WATERCARE_FORMAT = ">B"
SET_WATERCARE_FORMAT = ">BB"

_LOGGER = logging.getLogger(__name__)


class GeckoWatercareScheduleManager:
    """Watercare schedule manager."""


class GeckoGetWatercareModeProtocolHandler(GeckoPacketProtocolHandler):
    """Get watercare mode protocol handler class."""

    @staticmethod
    def get(seq: int, **kwargs: Any) -> GeckoGetWatercareModeProtocolHandler:
        """Generate a request."""
        return GeckoGetWatercareModeProtocolHandler(
            content=b"".join([GETWC_VERB, struct.pack(">B", seq)]),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def get_response(mode: int, **kwargs: Any) -> GeckoGetWatercareModeProtocolHandler:
        """Generate a watercare mode request response."""
        return GeckoGetWatercareModeProtocolHandler(
            content=b"".join(
                [
                    WCGET_VERB,
                    struct.pack(
                        GET_WATERCARE_FORMAT,
                        mode,
                    ),
                ]
            ),
            **kwargs,
        )

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle the verb."""
        return received_bytes.startswith((GETWC_VERB, WCGET_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        self.mode = None
        if received_bytes.startswith(GETWC_VERB):
            self._sequence = struct.unpack(">B", remainder)[0]
            return  # Stay in the handler list
        # Otherwise must be WCSET
        self.mode = struct.unpack(GET_WATERCARE_FORMAT, remainder)[0] % len(
            GeckoConstants.WATERCARE_MODE
        )
        self._should_remove_handler = True


class GeckoSetWatercareModeProtocolHandler(GeckoPacketProtocolHandler):
    """Set watercare mode protocol handler class."""

    @staticmethod
    def set(seq: int, mode: int, **kwargs: Any) -> GeckoSetWatercareModeProtocolHandler:
        """Generatge a watercare set command."""
        return GeckoSetWatercareModeProtocolHandler(
            content=b"".join(
                [SETWC_VERB, struct.pack(SET_WATERCARE_FORMAT, seq, mode)]
            ),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            **kwargs,
        )

    @staticmethod
    def set_response(mode: int, **kwargs: Any) -> GeckoSetWatercareModeProtocolHandler:
        """Generate a watercare set mode response."""
        return GeckoSetWatercareModeProtocolHandler(
            content=b"".join([WCSET_VERB, struct.pack(GET_WATERCARE_FORMAT, mode)]),
            **kwargs,
        )

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle the verb."""
        return received_bytes.startswith((SETWC_VERB, WCSET_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        self.mode = None
        if received_bytes.startswith(SETWC_VERB):
            self._sequence, self.mode = struct.unpack(SET_WATERCARE_FORMAT, remainder)
            self.mode %= len(GeckoConstants.WATERCARE_MODE)
            return  # Stay in the handler list
        # Otherwise must be WCSET
        self.mode = struct.unpack(GET_WATERCARE_FORMAT, remainder)[0]
        self._should_remove_handler = True


class GeckoGetWatercareScheduleListProtocolHandler(GeckoPacketProtocolHandler):
    """Get list of watercare schdeules."""

    @staticmethod
    def get(seq: int, **kwargs: Any) -> GeckoGetWatercareScheduleListProtocolHandler:
        """Generate a request."""
        return GeckoGetWatercareScheduleListProtocolHandler(
            content=b"".join([REQWC_VERB, struct.pack(">B", seq)]),
            timeout=GeckoConfig.PROTOCOL_TIMEOUT_IN_SECONDS,
            on_retry_failed=GeckoPacketProtocolHandler.default_retry_failed_handler,
            **kwargs,
        )

    @staticmethod
    def get_response(
        _schedule: GeckoWatercareScheduleManager, **kwargs: Any
    ) -> GeckoGetWatercareScheduleListProtocolHandler:
        """Generate the response."""
        return GeckoGetWatercareScheduleListProtocolHandler(
            content=b"".join(
                [
                    WCREQ_VERB,
                    b"\x00\x00",
                    b"\x00\x01\x00\x00\x06\x00\x00\x00\x00",
                    b"\x02\x01\x00\x01\x05\x06\x00\x12\x00"
                    b"\x03\x01\x00\x00\x06\x06\x00\x12\x00"
                    b"\x04\x01\x00\x01\x05\x00\x00\x00\x00",
                ]
            ),
            **kwargs,
        )

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle the verb."""
        return received_bytes.startswith((REQWC_VERB, WCREQ_VERB))

    def handle(self, received_bytes: bytes, _sender: tuple) -> None:
        """Handle the verb."""
        remainder = received_bytes[5:]
        if received_bytes.startswith(REQWC_VERB):
            self._sequence = struct.unpack(">B", remainder)[0]
            return  # Stay in the handler list
        # Otherwise must be WCREQ
        self._should_remove_handler = True


class GeckoWatercareErrorHandler(GeckoPacketProtocolHandler):
    """Watercare error handler."""

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle this verb."""
        return received_bytes.startswith(WCERR_VERB)

    def handle(self, _received_bytes: bytes, _sender: tuple) -> None:
        """Handle this."""



"""Gecko Bubble generator on Aux."""

from __future__ import annotations

from typing import TYPE_CHECKING

from .pump import GeckoPump

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade


class GeckoBubbleGenerator(GeckoPump):
    """Blowers are based on pumps."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the bubble generator."""
        super().__init__(facade, "Bubble Generator", "Aux")
        if self.is_available and "AuxAsBubbleGen" in self.facade.spa.accessors:
            self.set_availability(
                is_available=self.facade.spa.accessors["AuxAsBubbleGen"].value
            )



"""GeckoLogStruct - A class to manage the LogStruct for 'InYT v56'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoLogStruct:
    """Log Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the log struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the log struct class version."""
        return 56

    @property
    def begin(self) -> int:
        """Get the offset start."""
        return 256

    @property
    def end(self) -> int:
        """Get the offset end."""
        return 479

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return [
            "SOut1",
            "SOut2",
            "SOut3",
            "SOut4",
            "SOut5",
            "SOutHtr",
            "SOut6",
            "SOut7",
            "SOut8",
            "SOut9",
            "SOut10",
            "SOut11",
            "SOut12",
            "SDirect",
            "SDirect2",
        ]

    @property
    def all_device_keys(self) -> list[str]:
        """Get all device keys."""
        return [
            "P1",
            "P2",
            "P3",
            "P4",
            "P5",
            "BL",
            "CP",
            "O3",
            "L120",
            "MSTR_HEATER",
            "SLV_HEATER",
            "Waterfall",
            "LockMode",
            "DealerLockStatus",
            "DealerLockSeed",
            "DealerLockKey",
            "LI",
        ]

    @property
    def user_demand_keys(self) -> list[str]:
        """Get all user demand keys."""
        return [
            "UdP1",
            "UdP2",
            "UdP3",
            "UdP4",
            "UdP5",
            "UdBL",
            "UdL120",
            "UdLi",
            "UdWaterfall",
            "UdPumpTime",
            "UdQuietTime",
            "UdLightTime",
            "UdL120Time",
            "UdWaterFallTime",
        ]

    @property
    def error_keys(self) -> list[str]:
        """Get all error keys."""
        return [
            "AmbiantOHLevel2",
            "FLCErr",
            "FiltSuspendedByErr",
            "HeaterStuck",
            "KinPumpOff",
            "OverTemp",
            "P1HStuck",
            "P2HStuck",
            "RegOverHeat",
            "RelayStuck",
            "RhRegProbeErr",
            "SlaveAmbiantOHLevel2",
            "SlaveHLErr",
            "SlaveHtrStuck",
            "SlaveKinNoFloErr",
            "SlaveKinPumpOff",
            "SlaveMissingErr",
            "SlaveNoFloErr",
            "SlaveOverTemp",
            "SlaveP1HStuck",
            "SlaveP2HStuck",
            "SlaveRegOverHeat",
            "SlaveRegProbeErr",
            "SlaveRelayStuck",
            "SlaveThermFuseErr",
            "SlaveThermistanceErr",
            "TempNotValid",
            "ThermFuseErr",
            "ThermistanceErr",
        ]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "RhWaterTemp": GeckoTempStructAccessor(
                self.struct, "LogStructure/RealTimeTemp/RhWaterTemp", 317, None
            ),
            "Hours": GeckoByteStructAccessor(
                self.struct, "LogStructure/MiscStatus/Hours", 256, None
            ),
            "StickDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/MiscStatus/StickDetected", 274, 6, None
            ),
            "Menu": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/MiscStatus/Menu",
                319,
                None,
                [
                    "NORMAL",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "InstallerOpt",
                    "DealerOpt",
                    "AccessoryOpt",
                    "ConfigSelect",
                    "",
                    "StickBank",
                ],
                None,
                None,
                None,
            ),
            "QuietState": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/QuietState",
                257,
                None,
                ["NOT_SET", "DRAIN", "SOAK", "OFF"],
                None,
                None,
                "ALL",
            ),
            "UdP1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP1",
                259,
                0,
                ["OFF", "LO", "HI"],
                None,
                4,
                "ALL",
            ),
            "UdP2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP2",
                259,
                2,
                ["OFF", "LO", "HI"],
                None,
                4,
                "ALL",
            ),
            "UdP3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP3",
                259,
                4,
                ["OFF", "LO", "HI"],
                None,
                4,
                "ALL",
            ),
            "UdP4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP4",
                259,
                6,
                ["OFF", "LO", "HI"],
                None,
                4,
                "ALL",
            ),
            "UdP5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdP5",
                258,
                0,
                ["OFF", "HI"],
                None,
                2,
                "ALL",
            ),
            "UdBL": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdBL",
                258,
                1,
                ["OFF", "ON"],
                None,
                2,
                "ALL",
            ),
            "UdL120": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdL120",
                308,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "UdLi": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdLi",
                307,
                None,
                ["OFF", "HI"],
                None,
                None,
                "ALL",
            ),
            "UdWaterfall": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/UserDemands/UdWaterfall",
                363,
                None,
                ["OFF", "ON"],
                None,
                None,
                "ALL",
            ),
            "UdPumpTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdPumpTime", 303, "ALL"
            ),
            "UdQuietTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdQuietTime", 304, "ALL"
            ),
            "UdLightTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdLightTime", 305, "ALL"
            ),
            "UdL120Time": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdL120Time", 306, "ALL"
            ),
            "UdWaterFallTime": GeckoByteStructAccessor(
                self.struct, "LogStructure/UserDemands/UdWaterFallTime", 362, "ALL"
            ),
            "P1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P1",
                261,
                0,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P2",
                261,
                2,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P3",
                261,
                4,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P4",
                261,
                6,
                ["OFF", "HIGH", "LOW"],
                None,
                4,
                None,
            ),
            "P5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/P5",
                260,
                0,
                ["OFF", "HIGH"],
                None,
                2,
                None,
            ),
            "BL": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/BL",
                260,
                1,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "CP": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/CP",
                260,
                2,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "O3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/O3",
                260,
                3,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "L120": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/L120",
                260,
                4,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "MSTR_HEATER": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/MSTR_HEATER",
                260,
                5,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "SLV_HEATER": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/SLV_HEATER",
                260,
                6,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "Waterfall": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/Waterfall",
                260,
                7,
                ["OFF", "ON"],
                None,
                2,
                None,
            ),
            "LockMode": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/LockMode",
                310,
                None,
                ["UNLOCK", "PARTIAL", "FULL"],
                None,
                None,
                "ALL",
            ),
            "DealerLockStatus": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/DeviceStatus/DealerLockStatus",
                364,
                None,
                ["UNLOCK", "PARTIAL", "FULL"],
                None,
                None,
                None,
            ),
            "DealerLockSeed": GeckoWordStructAccessor(
                self.struct, "LogStructure/DeviceStatus/DealerLockSeed", 365, None
            ),
            "DealerLockKey": GeckoWordStructAccessor(
                self.struct, "LogStructure/DeviceStatus/DealerLockKey", 367, "ALL"
            ),
            "FilterAccess": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/FilterAccess",
                262,
                0,
                ["INTERNAL", "REMOTE"],
                None,
                2,
                "ALL",
            ),
            "RemoteFiltAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteFiltAction",
                263,
                None,
                ["IDLE", "STOP", "START", "NEW", "ACTIVE"],
                None,
                None,
                "ALL",
            ),
            "RemoteFiltDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteFiltDur", 264, "ALL"
            ),
            "RemoteFiltDurPerDay": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteFiltDurPerDay",
                266,
                "ALL",
            ),
            "EconomyAccess": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/EconomyAccess",
                262,
                2,
                ["INTERNAL", "REMOTE"],
                None,
                2,
                "ALL",
            ),
            "RemoteEconAction": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteEconAction",
                267,
                None,
                ["IDLE", "STOP", "START", "NEW", "ACTIVE"],
                None,
                None,
                "ALL",
            ),
            "RemoteEconDur": GeckoTimeStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteEconDur", 268, "ALL"
            ),
            "RemoteConfigIndex": GeckoByteStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteConfigIndex", 270, "ALL"
            ),
            "RemoteNbOfPhases": GeckoByteStructAccessor(
                self.struct, "LogStructure/RemoteCommands/RemoteNbOfPhases", 271, "ALL"
            ),
            "RemoteBreakerIndex": GeckoByteStructAccessor(
                self.struct,
                "LogStructure/RemoteCommands/RemoteBreakerIndex",
                272,
                "ALL",
            ),
            "Clean": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/Clean", 273, 0, None
            ),
            "Purge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/Purge", 273, 2, None
            ),
            "FiltSuspendByUD": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/FiltSuspendByUD", 273, 3, None
            ),
            "FiltSuspendedByOT": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/FiltSuspendedByOT", 273, 4, None
            ),
            "FiltSuspendedByErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/FilterStatus/FiltSuspendedByErr",
                273,
                5,
                None,
            ),
            "CPOT": GeckoBoolStructAccessor(
                self.struct, "LogStructure/FilterStatus/CPOT", 274, 2, None
            ),
            "SwmPurgeSusp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmPurgeSusp", 282, 3, None
            ),
            "SwmPurge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmPurge", 282, 5, None
            ),
            "SwmActive": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmActive", 282, 6, None
            ),
            "SwmRisk": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SWMStatus/SwmRisk",
                313,
                None,
                ["NO", "LO", "MED", "HI", "EXTREME"],
                None,
                None,
                None,
            ),
            "SlaveSwmPurge": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SlaveSwmPurge", 353, 5, None
            ),
            "SlaveSwmActive": GeckoBoolStructAccessor(
                self.struct, "LogStructure/SWMStatus/SlaveSwmActive", 353, 6, None
            ),
            "SwmAdc": GeckoWordStructAccessor(
                self.struct, "LogStructure/SWMStatus/SwmAdc", 355, None
            ),
            "OverTemp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/OverTemp", 274, 1, None
            ),
            "RealSetPointG": GeckoTempStructAccessor(
                self.struct, "LogStructure/RegulationStatus/RealSetPointG", 275, None
            ),
            "DisplayedTempG": GeckoTempStructAccessor(
                self.struct, "LogStructure/RegulationStatus/DisplayedTempG", 277, None
            ),
            "Heating": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/Heating",
                260,
                5,
                ["", "Heating", "Heating", "Heating"],
                None,
                4,
                None,
            ),
            "TempNotValid": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/TempNotValid", 279, 2, None
            ),
            "ExtProbeDetected": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/ExtProbeDetected",
                279,
                6,
                None,
            ),
            "CheckFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/CheckFlo", 280, 2, None
            ),
            "ProgEconActive": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/RegulationStatus/ProgEconActive",
                281,
                1,
                None,
            ),
            "EconActive": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/EconActive", 281, 2, "ALL"
            ),
            "SlaveOverTemp": GeckoBoolStructAccessor(
                self.struct, "LogStructure/RegulationStatus/SlaveOverTemp", 352, 1, None
            ),
            "ThermFuseErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/ThermFuseErr", 274, 3, None
            ),
            "ThermistanceErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/ThermistanceErr", 282, 0, None
            ),
            "AmbiantOHLevel2": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/AmbiantOHLevel2", 282, 1, None
            ),
            "KinPumpOff": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/KinPumpOff", 283, 2, None
            ),
            "RegOverHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/RegOverHeat", 283, 3, None
            ),
            "P1HStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/P1HStuck", 284, 3, None
            ),
            "P2HStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/P2HStuck", 284, 4, None
            ),
            "HeaterStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/HeaterStuck", 284, 5, None
            ),
            "RelayStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/RelayStuck", 284, 6, None
            ),
            "SlaveP1HStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveP1HStuck", 350, 3, None
            ),
            "SlaveP2HStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveP2HStuck", 350, 4, None
            ),
            "SlaveHtrStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveHtrStuck", 350, 5, None
            ),
            "SlaveRelayStuck": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveRelayStuck", 350, 6, None
            ),
            "SlaveKinPumpOff": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveKinPumpOff", 351, 2, None
            ),
            "SlaveRegOverHeat": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveRegOverHeat", 351, 3, None
            ),
            "SlaveThermFuseErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/SlaveThermFuseErr",
                352,
                3,
                None,
            ),
            "SlaveThermistanceErr": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/SlaveThermistanceErr",
                353,
                0,
                None,
            ),
            "SlaveAmbiantOHLevel2": GeckoBoolStructAccessor(
                self.struct,
                "LogStructure/ErrorMessages/SlaveAmbiantOHLevel2",
                353,
                1,
                None,
            ),
            "SlaveMissingErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/ErrorMessages/SlaveMissingErr", 354, 0, None
            ),
            "RhFloDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhFloDetected", 280, 0, None
            ),
            "RhHwHL": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHwHL", 283, 0, None
            ),
            "RhRegProbeErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhRegProbeErr", 283, 1, None
            ),
            "RhRegSlope": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhRegSlope", 283, 4, None
            ),
            "RhHrKinNoFlo": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/RhHrKinNoFlo", 309, 0, None
            ),
            "RhNoFloXTries": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/InthermStatus/RhNoFloXTries",
                309,
                1,
                ["", "RhNoFloXTries", "RhNoFloXTries", "RhNoFloXTries"],
                None,
                4,
                None,
            ),
            "FLCErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/FLCErr", 309, 3, None
            ),
            "inTCipDelay": GeckoWordStructAccessor(
                self.struct, "LogStructure/InthermStatus/inTCipDelay", 311, None
            ),
            "SlaveFloDetected": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/SlaveFloDetected", 314, 0, None
            ),
            "SlaveKinNoFloErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/SlaveKinNoFloErr", 315, 0, None
            ),
            "SlaveNoFloErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/SlaveNoFloErr", 315, 1, None
            ),
            "SlaveHLErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/SlaveHLErr", 351, 0, None
            ),
            "SlaveRegProbeErr": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/SlaveRegProbeErr", 351, 1, None
            ),
            "SlaveRegSlope": GeckoBoolStructAccessor(
                self.struct, "LogStructure/InthermStatus/SlaveRegSlope", 351, 4, None
            ),
            "PackBootID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootID", 285, None
            ),
            "PackBootRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRev", 287, None
            ),
            "PackBootRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackBootRel", 288, None
            ),
            "PackType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackType",
                289,
                None,
                [
                    "Unknown",
                    "inXE",
                    "MasIBC",
                    "MIA",
                    "DJS4",
                    "inClear",
                    "inXM",
                    "K600",
                    "inTerface",
                    "inTouch",
                    "inYT",
                ],
                None,
                None,
                None,
            ),
            "PackMemRange": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackMemRange",
                290,
                0,
                ["16K", "32K", "48K", "64K"],
                None,
                4,
                None,
            ),
            "PackRegion": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackRegion",
                290,
                2,
                ["UL", "CE"],
                None,
                2,
                None,
            ),
            "PackCoreID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreID", 291, None
            ),
            "PackCoreRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRev", 293, None
            ),
            "PackCoreRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackCoreRel", 294, None
            ),
            "PackConfigLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfigLib", 295, None
            ),
            "PackStatusLib": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackStatusLib", 296, None
            ),
            "PackConfID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfID", 297, None
            ),
            "PackConfRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfRev", 299, None
            ),
            "PackConfRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/PackConfRel", 300, None
            ),
            "PackNumberOfConf": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/PackNumberOfConf", 301, None
            ),
            "PackLogTrig": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/PackLogTrig",
                316,
                None,
                ["Restricted", "Full"],
                None,
                2,
                "ALL",
            ),
            "KeypadType": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/PackInfo/KeypadType",
                357,
                None,
                [
                    "K200",
                    "K400",
                    "K85",
                    "K8",
                    "K4",
                    "K5",
                    "K600LE",
                    "K100",
                    "K800",
                    "",
                    "",
                    "",
                    "K600HE",
                    "K300",
                    "INVALID",
                ],
                None,
                None,
                None,
            ),
            "KeypadID": GeckoWordStructAccessor(
                self.struct, "LogStructure/PackInfo/KeypadID", 358, None
            ),
            "KeypadRev": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/KeypadRev", 360, None
            ),
            "KeypadRel": GeckoByteStructAccessor(
                self.struct, "LogStructure/PackInfo/KeypadRel", 361, None
            ),
            "SOut1": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveHCOutputConfig/SOut1",
                320,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveHCOutputConfig/SOut2",
                321,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut3": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveHCOutputConfig/SOut3",
                322,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut4": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveHCOutputConfig/SOut4",
                323,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut5": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveHCOutputConfig/SOut5",
                324,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOutHtr": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveHCOutputConfig/SOutHtr",
                334,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "", "", "", "HTR"],
                None,
                None,
                "ALL",
            ),
            "SOut1Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveHCOutputConfig/SOut1Cur", 335, "ALL"
            ),
            "SOut2Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveHCOutputConfig/SOut2Cur", 336, "ALL"
            ),
            "SOut3Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveHCOutputConfig/SOut3Cur", 337, "ALL"
            ),
            "SOut4Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveHCOutputConfig/SOut4Cur", 338, "ALL"
            ),
            "SOut5Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveHCOutputConfig/SOut5Cur", 339, "ALL"
            ),
            "SOutHtrCur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveHCOutputConfig/SOutHtrCur", 349, "ALL"
            ),
            "SOut6": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut6",
                325,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut7": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut7",
                326,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut8": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut8",
                327,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut9": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut9",
                328,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut10": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut10",
                329,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut11": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut11",
                330,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SOut12": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SOut12",
                331,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "",
                    "",
                    "Fan",
                    "",
                    "",
                    "FullOn",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "",
                    "Waterfall",
                ],
                None,
                None,
                "ALL",
            ),
            "SDirect": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SDirect",
                332,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "CP"],
                None,
                None,
                "ALL",
            ),
            "SDirect2": GeckoEnumStructAccessor(
                self.struct,
                "LogStructure/SlaveLCOutputConfig/SDirect2",
                333,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "CP"],
                None,
                None,
                "ALL",
            ),
            "SOut6Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut6Cur", 340, "ALL"
            ),
            "SOut7Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut7Cur", 341, "ALL"
            ),
            "SOut8Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut8Cur", 342, "ALL"
            ),
            "SOut9Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut9Cur", 343, "ALL"
            ),
            "SOut10Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut10Cur", 344, "ALL"
            ),
            "SOut11Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut11Cur", 345, "ALL"
            ),
            "SOut12Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SOut12Cur", 346, "ALL"
            ),
            "SDirectCur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SDirectCur", 347, "ALL"
            ),
            "SDirect2Cur": GeckoByteStructAccessor(
                self.struct, "LogStructure/SlaveLCOutputConfig/SDirect2Cur", 348, "ALL"
            ),
        }



"""GeckoUdpProtocolHandler base class."""

from __future__ import annotations

import asyncio
import logging
import time
from typing import TYPE_CHECKING, Any

from geckolib.config import GeckoConfig, config_sleep

if TYPE_CHECKING:
    from .async_udp_protocol import GeckoAsyncUdpProtocol


from abc import ABC, abstractmethod

_LOGGER = logging.getLogger(__name__)


class GeckoUdpProtocolHandler(ABC):
    """
    Protocol handlers class.

    Manage both sides of a specific conversation part with
    a remote end.

    The protocol is either initiated by a client or by a server, but either
    way a query should always met with a response from the remote end

    Both sides may instantiate listening handlers which will deal with
    unsolicited requests from remote clients and will respond so that the
    remote end knows the request was received and they may also send a query
    to the remote end, expecting a response to confirm receipt.

    A message sent will either make it to the destination, or it won't. Since
    this protocol is built on UDP, we have to cook our own timeout/retry
    mechanism to ensure delivery ... oddly this is precisely what TCP already
    does, no idea why this wasn't considered but hey-ho, this is all a bit of
    fun anyway!

    The base protocol handler will manage the lifetime of the handlers within
    the socket, and mechanisms are in place to allow retries to be handled
    with failure exit points available to allow clients to make class decisions
    or instance decisions, either by overridden methods, or by instance handlers.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the protocol handler class."""
        # Send functionality
        self._send_bytes = kwargs.get("send_bytes")
        self.last_destination = None

        # Receive functionality
        self._async_on_handled = kwargs.get("async_on_handled")

        # Lifetime functionality
        self._start_time = time.monotonic()
        self._timeout_in_seconds: float = kwargs.get("timeout", 0)
        self._on_retry_failed = kwargs.get("on_retry_failed")
        self._should_remove_handler = False

    ##########################################################################
    #
    #                         SEND FUNCTIONALITY
    #
    ##########################################################################
    @property
    def send_bytes(self) -> bytes:
        """
        The bytes to send to the remote end.

        Either uses the class instancedata _send_bytes or can be overridden
        in a base class.
        """
        if self._send_bytes is None:
            raise NotImplementedError
        return self._send_bytes

    @property
    def timeout_in_seconds(self) -> float:
        """The timeout in seconds."""
        return self._timeout_in_seconds

    ##########################################################################
    #
    #                        RECEIVE FUNCTIONALITY
    #
    ##########################################################################

    @abstractmethod
    def can_handle(self, received_bytes: bytes, sender: tuple) -> bool:
        """
        Check if you can handle these bytes.

        If you return True, then your handle method will get called and no
        more handlers will be given a chance to process this data. If you
        return False then the search for a suitable handler will continue
        """
        return False

    @abstractmethod
    def handle(self, received_bytes: bytes, sender: tuple) -> None:
        """
        Handle this data.

        This will only be called if you returned True from the `can_handle`
        function. If you wish to remove this handler from the system, then
        you should set the `should_remove_handler member.
        """

    ##########################################################################
    #
    #                           ASYNC FUNCTIONALITY
    #
    ##########################################################################

    async def async_handle(self, received_bytes: bytes, sender: tuple) -> None:
        """
        Handle this data.

        This will only be called if you returned True from the `can_handle`
        function. If you wish to remove this handler from the system, then you
        should set the `should_remove_handler` member.
        """
        self.handle(received_bytes, sender)

    async def async_handled(self, sender: tuple) -> None:
        """Handle the data in the base class. Derivatives must call this."""
        self._reset_timeout()
        if self._async_on_handled is not None:
            await self._async_on_handled(self, sender)

    async def wait_for_response(
        self,
        protocol: GeckoAsyncUdpProtocol,
        packet_timeout: int = GeckoConfig.PAUSE_BETWEEN_RETRIES_IN_SECONDS,
    ) -> bool:
        """
        Wait for a response that this command can handle.

        This function doesn't respect the _async_on_handled functionality since its
        use is for inline async stuff.
        """
        assert self.timeout_in_seconds > 0  # noqa: S101
        _LOGGER.debug("Wait for response for %s", self)
        try:
            while True:
                try:
                    await config_sleep(None, "Async UDP handler - wait for response")

                    async with asyncio.timeout(packet_timeout):
                        await protocol.queue.wait()

                    peeked_data = protocol.queue.peek()
                    if peeked_data is None:
                        continue

                    data, sender = peeked_data
                    if self.can_handle(data, sender):
                        protocol.queue.pop()
                        await self.async_handle(data, sender)
                        self._reset_timeout()
                        return False

                except TimeoutError:
                    _LOGGER.debug(
                        "TIMEOUT: Packet timeout for %s threw TimeoutError (%fs)",
                        self,
                        packet_timeout,
                    )
                    return True

        except Exception:
            _LOGGER.exception("wait_for_response loop for %r caught exception", self)
            return True
        finally:
            _LOGGER.debug("wait_for_response loop for %r terminated", self)

    def _can_handle(
        self, _protocol: GeckoAsyncUdpProtocol, received_bytes: bytes, sender: tuple
    ) -> bool:
        """Provide for derived classes to overide."""
        return self.can_handle(received_bytes, sender)

    async def consume(self, protocol: GeckoAsyncUdpProtocol) -> None:
        """Async coroutine to handle datagram."""
        try:
            assert self.timeout_in_seconds == 0  # noqa: S101
            while True:
                await config_sleep(None, f"Async UDP handler - consume for {self}")
                await protocol.queue.wait()

                peeked_data = protocol.queue.peek()
                if peeked_data is None:
                    continue

                data, sender = peeked_data
                if self._can_handle(protocol, data, sender):
                    protocol.queue.pop()
                    await self.async_handle(data, sender)
                    await self.async_handled(sender)

                if self.should_remove_handler:
                    _LOGGER.debug("%s will be removed, consume loop terminating", self)
                    break

        except asyncio.CancelledError:
            _LOGGER.debug("Consume loop for %r cancelled", self)
            raise
        except Exception:
            _LOGGER.exception("Consume loop for %r caught exception", self)
            raise
        finally:
            _LOGGER.debug("Consume loop for %r terminated", self)

    ##########################################################################
    #
    #                         LIFETIME MANAGEMENT
    #
    ##########################################################################
    @property
    def age(self) -> float:
        """Get the age of this handler."""
        return time.monotonic() - self._start_time

    @property
    def should_remove_handler(self) -> bool:
        """Get the should remove handler property."""
        return self._should_remove_handler

    def _reset_timeout(self) -> None:
        self._start_time = time.monotonic()

    @staticmethod
    def default_retry_failed_handler(
        handler: GeckoUdpProtocolHandler, _protocol: GeckoAsyncUdpProtocol
    ) -> None:
        """Handle retry failure."""
        _LOGGER.debug("Default retry failed handler for %r being used", handler)
        handler._should_remove_handler = True

    # Pythonic methods
    def __repr__(self) -> str:
        """Get string representation of this class."""
        return (
            f"{self.__class__.__name__}(send_bytes={self._send_bytes!r},"
            f" age={self.age},"
            f" should_remove_handler={self.should_remove_handler},"
            f" timeout={self.timeout_in_seconds}s"
            f")"
        )



"""GeckoAsyncSpaDescriptor class."""

import logging

from .const import GeckoConstants

logger = logging.getLogger(__name__)


class GeckoAsyncSpaDescriptor:
    """A descriptor class for spas that have been discovered on the network."""

    def __init__(
        self,
        spa_identifier: bytes,
        spa_name: str,
        sender: tuple,
    ) -> None:
        """Initialize the descriptor."""
        self.identifier = spa_identifier
        self.name = spa_name
        self.ipaddress, self.port = sender

    @property
    def identifier_as_string(self) -> str:
        """
        The spa identifier as a string.

        Useful for storing in configuration and
        then passing as spa_identifier to the facade
        """
        return self.identifier.decode(GeckoConstants.MESSAGE_ENCODING)

    @property
    def destination(self) -> tuple:
        """The destination of this descriptor."""
        return (self.ipaddress, self.port)

    def __repr__(self) -> str:
        """Get then string representation."""
        return f"{self.name}({self.identifier_as_string}) [{self.destination}]"



"""Pack Module Initialization."""

from geckolib.driver.accessor import (
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)
from geckolib.driver.async_spastruct import GeckoAsyncStructure

__all__ = [
    "GeckoAsyncStructure",
    "GeckoBoolStructAccessor",
    "GeckoByteStructAccessor",
    "GeckoEnumStructAccessor",
    "GeckoStructAccessor",
    "GeckoTempStructAccessor",
    "GeckoTimeStructAccessor",
    "GeckoWordStructAccessor",
]



"""GeckoButton class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoButton(GeckoAutomationFacadeBase):
    """A button can be pressed ... that's it."""

    def __init__(self, facade: GeckoAsyncFacade, name: str, keypad: int) -> None:
        """Inirtialize the button class."""
        super().__init__(facade, name, f"KEYPAD{keypad}")
        self._keypad: int = keypad

    async def async_press(self) -> None:
        """Press the button."""
        await self.facade.spa.async_press(self._keypad)



"""
Abstract curses display class for use in asyncio app.

Thanks to https://gist.github.com/davesteele/8838f03e0594ef11c89f77a7bca91206
"""

import _curses
import asyncio
import curses
import inspect
import logging
from abc import ABC, abstractmethod
from collections.abc import Callable
from curses import ERR, KEY_MOUSE, KEY_RESIZE, curs_set, getmouse
from typing import Any, cast

from geckolib import GeckoAsyncTaskMan

_LOGGER = logging.getLogger(__name__)


class Button:
    """A button wrapper."""

    def __init__(  # noqa: PLR0913
        self,
        stdscr: _curses.window,
        y: int,
        x: int,
        text: str | list[str],
        char: str | None,
        command: Callable | tuple[Callable, Any] | None,
    ) -> None:
        """Initialize the button."""
        self.stdscr = stdscr
        self.y = y
        self.x = x
        self.text: list[str] = text if isinstance(text, list) else [text]
        self.char = char
        self.command = command

        self.width = max([len(t) for t in self.text]) + 4
        self.height = len(self.text) + 2
        self.window = None

    def _create_window(self) -> None:
        if self.window is None:
            self.window = self.stdscr.subwin(self.height, self.width, self.y, self.x)

    def make(self, attr: int = curses.A_NORMAL) -> None:
        """Make the curses objects."""
        self._create_window()
        if self.window is None:
            return
        self.window.erase()
        self.window.box()
        for idx, txt in enumerate(self.text):
            self.window.addstr(idx + 1, int((self.width - len(txt)) / 2), txt, attr)
        self.window.refresh()

    def is_in_button(self, y: int, x: int) -> bool:
        """Test if point is in the button."""
        self._create_window()
        if self.window is None:
            return False
        return self.window.enclose(y, x)

    def move(self, new_y: int, new_x: int) -> None:
        """Move the button to a new location."""
        self.y = new_y
        self.x = new_x
        if self.window is None:
            return
        self.window.mvderwin(new_y, new_x)
        self.window.mvwin(new_y, new_x)

    def __str__(self) -> str:
        """Stringize the button."""
        return (
            f"Button ({self.text} with '{self.char}') at ({self.y},"
            f" {self.x}, {self.height}, {self.width})"
        )


class AbstractDisplay(ABC):
    """Abstract display class."""

    def __init__(self, stdscr: _curses.window) -> None:
        """Initialize the class."""
        self.stdscr = stdscr
        self.done_event = asyncio.Event()
        self.queue = asyncio.Queue(5)

        # Enable mouse events
        curses.mousemask(curses.ALL_MOUSE_EVENTS | curses.REPORT_MOUSE_POSITION)

        # https://stackoverflow.com/questions/56300134/how-to-enable-mouse-movement-events-in-curses#64809709
        print("\033[?1003h")  # enable mouse tracking with the XTERM API  # noqa: T201
        # https://invisible-island.net/xterm/ctlseqs/ctlseqs.html#h2-Mouse-Tracking

        # Temp waiting to get replaced
        self._commands = {}
        self._buttons: list[Button] = []

        self._last_x = 0
        self._last_y = 0
        self._last_button = None

    @abstractmethod
    def make_display(self) -> None:
        """Make a display."""

    async def execute(self, func: Any) -> None:
        """Execute a function, or unpack and execute a function tuple."""
        try:
            if func is None:
                return
            if isinstance(func, Button):
                await self.execute(cast(Button, func).command)
                return
            if inspect.iscoroutinefunction(func):
                await func()
            elif isinstance(func, tuple):
                func, *parms = func
                if inspect.iscoroutinefunction(func):
                    await func(*parms)
                else:
                    func(*parms)
            else:
                func()
        except Exception:
            _LOGGER.exception("Exception during execute %s", func)
            raise

    async def handle_char(self, char: int) -> None:
        """Handle a command character."""
        cmd = chr(char)
        if cmd in self._commands:
            await self.execute(self._commands[cmd])

    def _get_button(self, y: int, x: int) -> Button | None:
        for button in self._buttons:
            if button.is_in_button(y, x):
                return button
        return None

    async def handle_mouse(self, mouse: tuple[int, int, int, int, int]) -> None:
        """Handle the mouse events."""
        _id, x, y, z, bstate = mouse
        self._last_x = x
        self._last_y = y
        button = self._get_button(y, x)
        if button is not None:
            if (
                bstate & curses.BUTTON1_PRESSED == curses.BUTTON1_PRESSED
                or bstate & curses.BUTTON1_CLICKED == curses.BUTTON1_CLICKED
                or bstate & curses.BUTTON1_DOUBLE_CLICKED
                == curses.BUTTON1_DOUBLE_CLICKED
            ):
                await self.execute(button.command)

            if button is not self._last_button:
                button.make(curses.A_BOLD)
                self._last_button = button

        elif self._last_button is not None:
            self._last_button.make()
            self._last_button = None

    def refresh(self) -> None:
        """Refesh the display."""
        self._commands = {}
        self._buttons = []
        self._last_button = None
        self.stdscr.erase()
        self.make_display()
        for button in self._buttons:
            button.make(
                curses.A_BOLD
                if button.is_in_button(self._last_y, self._last_x)
                else curses.A_NORMAL
            )
        self.stdscr.refresh()

    def set_exit(self) -> None:
        """Indicate we can exit."""
        self.done_event.set()

    async def enqueue_input(self) -> None:
        """Get input and queue it up."""
        try:
            while not self.done_event.is_set():
                char = self.stdscr.getch()
                await self.queue.put(char)
        except asyncio.CancelledError:
            _LOGGER.debug("CUI enqueue loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("CUI enqueue loop caught exception")
            raise

    async def process_input(self) -> None:
        """Get queue data and process it."""
        try:
            while not self.done_event.is_set():
                try:
                    char = await self.queue.get()
                    if char == ERR:
                        # Do nothing and let the loop continue without sleeping
                        pass
                    elif char == KEY_RESIZE:
                        self.refresh()
                    elif char == KEY_MOUSE:
                        await self.handle_mouse(getmouse())
                    else:
                        await self.handle_char(char)
                    self.queue.task_done()

                except asyncio.CancelledError:
                    _LOGGER.debug("Input loop cancelled")
                    raise

                except Exception:
                    _LOGGER.exception("Exception in input loop, ignored and continuing")

        finally:
            _LOGGER.debug("Input loop finished")

    async def run(self, taskman: GeckoAsyncTaskMan) -> None:
        """Run the display class."""
        curs_set(0)
        self.stdscr.nodelay(True)  # noqa: FBT003

        self.refresh()

        taskman.add_task(self.enqueue_input(), "Input gather", "CUI")
        taskman.add_task(self.process_input(), "Process input", "CUI")
        await self.done_event.wait()
        taskman.cancel_key_tasks("CUI")

    def add_text_box(
        self, y: int, x: int, text: str | list[str]
    ) -> tuple[int, int, int, int]:
        """Add a text box."""
        if isinstance(text, str):
            text = [text]
        width = max([len(t) for t in text]) + 2
        height = len(text)

        self.stdscr.addstr(y, x, f"┌{'─' * width}┐")
        for idx, line in enumerate(text):
            self.stdscr.addstr(y + 1 + idx, x, f"│ {line} │")
        self.stdscr.addstr(y + 1 + height, x, f"└{'─' * width}┘")
        return (y, x, height + 2, width + 2)

    def add_button(
        self,
        y: int,
        x: int,
        text: str | list[str],
        char: str | None,
        command: Callable | tuple[Callable, Any] | None,
    ) -> Button:
        """Add a button to the window."""
        button = Button(self.stdscr, y, x, text, char, command)
        self._buttons.append(button)
        if char is not None:
            self._commands[char] = button
        return button

    def add_line(self, y: int, x: int, text: str) -> None:
        """Add a text line."""
        self.stdscr.addstr(y, x, text)



"""Gecko RFERR handlers."""

from __future__ import annotations

import logging
from datetime import UTC, datetime
from typing import Any

from .packet import GeckoPacketProtocolHandler

RFERR_VERB = b"RFERR"

_LOGGER = logging.getLogger(__name__)


class GeckoRFErrProtocolHandler(GeckoPacketProtocolHandler):
    """Handle RFERR."""

    @staticmethod
    def response(**kwargs: Any) -> GeckoRFErrProtocolHandler:
        """Generate a response."""
        return GeckoRFErrProtocolHandler(content=RFERR_VERB, **kwargs)

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the class."""
        super().__init__(**kwargs)
        self._error_count: int = 0
        self._last_error_at: datetime | None = None

    def can_handle(self, received_bytes: bytes, _sender: tuple) -> bool:
        """Can we handle this."""
        return received_bytes.startswith(RFERR_VERB)

    def handle(self, _received_bytes: bytes, _sender: tuple) -> None:
        """Handle RFERR."""
        self._error_count += 1
        self._last_error_at = datetime.now(tz=UTC)
        _LOGGER.debug(
            "RF error #%d, intouch2 EN module cannot communicate with spa (CO) module",
            self.total_error_count,
        )

    @property
    def total_error_count(self) -> int:
        """Return total number of RFErr messages that this handler has processed."""
        return self._error_count

    @property
    def last_error_at(self) -> datetime | None:
        """Return the last time an RFErr occurred, or None if never."""
        return self._last_error_at



"""GeckoConfigStruct - A class to manage the ConfigStruct for 'InXE-64K v57'."""  # noqa: N999

from . import (
    GeckoAsyncStructure,
    GeckoBoolStructAccessor,
    GeckoByteStructAccessor,
    GeckoEnumStructAccessor,
    GeckoStructAccessor,
    GeckoTempStructAccessor,
    GeckoTimeStructAccessor,
    GeckoWordStructAccessor,
)


class GeckoConfigStruct:
    """Config Struct Class."""

    def __init__(self, struct_: GeckoAsyncStructure) -> None:
        """Initialize the config struct class."""
        self.struct = struct_

    @property
    def version(self) -> int:
        """Get the config struct class version."""
        return 57

    @property
    def output_keys(self) -> list[str]:
        """Output keys property."""
        return ["Out1", "Out2", "Out3", "Out4", "Out5", "OutHtr", "Direct", "OutLi"]

    @property
    def accessors(self) -> dict[str, GeckoStructAccessor]:
        """The structure accessors."""
        return {
            "ConfigNumber": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/ConfigNumber", 0, "ALL"
            ),
            "Out1": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out1",
                12,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "",
                    "",
                    "",
                    "FullOn",
                ],
                None,
                None,
                "ALL",
            ),
            "Out2": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out2",
                13,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "",
                    "",
                    "",
                    "FullOn",
                ],
                None,
                None,
                "ALL",
            ),
            "Out3": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out3",
                14,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "",
                    "",
                    "",
                    "FullOn",
                ],
                None,
                None,
                "ALL",
            ),
            "Out4": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out4",
                15,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "",
                    "",
                    "",
                    "FullOn",
                ],
                None,
                None,
                "ALL",
            ),
            "Out5": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/Out5",
                16,
                None,
                [
                    "NA",
                    "P1H",
                    "P1L",
                    "P2H",
                    "P2L",
                    "P3H",
                    "P3L",
                    "P4H",
                    "P4L",
                    "P5",
                    "BLO",
                    "CP",
                    "O3",
                    "L120",
                    "HTR",
                    "",
                    "",
                    "",
                    "",
                    "FullOn",
                ],
                None,
                None,
                "ALL",
            ),
            "OutHtr": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/HCOutputConfig/OutHtr",
                18,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "", "", "", "HTR"],
                None,
                None,
                "ALL",
            ),
            "Out1Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out1Cur", 36, "ALL"
            ),
            "Out2Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out2Cur", 37, "ALL"
            ),
            "Out3Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out3Cur", 38, "ALL"
            ),
            "Out4Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out4Cur", 39, "ALL"
            ),
            "Out5Cur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/Out5Cur", 40, "ALL"
            ),
            "OutHtRCur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/HCOutputConfig/OutHtRCur", 42, "ALL"
            ),
            "Direct": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LCOutputConfig/Direct",
                17,
                None,
                ["NA", "", "", "", "", "", "", "", "", "", "", "CP"],
                None,
                None,
                "ALL",
            ),
            "DirectCur": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LCOutputConfig/DirectCur", 41, "ALL"
            ),
            "OutLi": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/LVOutputConfig/OutLi",
                79,
                None,
                ["", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "LI"],
                None,
                None,
                None,
            ),
            "LightInts": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/LVOutputConfig/LightInts", 80, None
            ),
            "PumpTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/PumpTimeOut", 54, "ALL"
            ),
            "LightTimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/LightTimeOut", 55, "ALL"
            ),
            "L120TimeOut": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/TimeOut/L120TimeOut", 56, "ALL"
            ),
            "L120Timer": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/TimeOut/L120Timer",
                81,
                0,
                ["Shared", "Own"],
                None,
                2,
                "ALL",
            ),
            "FiltInterface": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/FiltInterface",
                32,
                None,
                ["PurgeOnly", "FiltCP", "FiltP1", "FiltP1DurOnly"],
                None,
                None,
                "ALL",
            ),
            "CpUsage": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/CpUsage",
                27,
                None,
                ["STANDARD", "ALWAYS_ON"],
                None,
                None,
                "ALL",
            ),
            "OtOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/OtOption",
                57,
                None,
                ["Disabled", "AlwaysEnabled", "WithSPOver95F"],
                None,
                None,
                "ALL",
            ),
            "PurgeSpeed": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/FilterConfig/PurgeSpeed",
                81,
                2,
                ["Lo", "Hi"],
                None,
                2,
                "ALL",
            ),
            "OTTriggerG": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/AdvanceFilterConfig/OTTriggerG", 58, "ALL"
            ),
            "CpOnTimeDuringOT": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/CpOnTimeDuringOT",
                59,
                "ALL",
            ),
            "CpOffTimeDuringOT": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/CpOffTimeDuringOT",
                60,
                "ALL",
            ),
            "FiltOnTimeDuringOT": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltOnTimeDuringOT",
                61,
                "ALL",
            ),
            "FiltSuspendTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/FiltSuspendTime",
                62,
                "ALL",
            ),
            "DrainMode": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/AdvanceFilterConfig/DrainMode",
                78,
                None,
                ["NA", "P1", "CP"],
                None,
                None,
                "ALL",
            ),
            "O3Usage": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Usage",
                28,
                None,
                ["Filter", "Always"],
                None,
                None,
                "ALL",
            ),
            "O3Pump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Pump",
                29,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "O3Type": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/O3Config/O3Type",
                30,
                None,
                ["Standard", "Toggle"],
                None,
                None,
                "ALL",
            ),
            "O3SuspendTime": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/O3Config/O3SuspendTime", 63, "ALL"
            ),
            "SetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/SetpointG",
                1,
                "ALL",
            ),
            "TempUnits": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/TempUnits",
                33,
                None,
                ["F", "C"],
                None,
                None,
                "ALL",
            ),
            "HeaterPump": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/HeaterPump",
                31,
                None,
                ["CP", "P1"],
                None,
                None,
                "ALL",
            ),
            "FlowDetector": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/FlowDetector",
                25,
                None,
                ["inFlo", "NotInstalled"],
                None,
                None,
                "ALL",
            ),
            "ProbeLocation": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/ProbeLocation",
                81,
                3,
                ["IntoPiping", "IntoTub"],
                None,
                2,
                "ALL",
            ),
            "CooldownTime": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/CooldownTime",
                35,
                "ALL",
            ),
            "MinSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/MinSetpointG",
                66,
                "ALL",
            ),
            "MaxSetpointG": GeckoTempStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/MaxSetpointG",
                68,
                "ALL",
            ),
            "EconType": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/EconType",
                70,
                None,
                ["Standard", "Night"],
                None,
                None,
                "ALL",
            ),
            "NoHeatPeriod": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/WaterTemperatureControl/NoHeatPeriod",
                77,
                "ALL",
            ),
            "MaxNumberOfPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/MaxNumberOfPhases", 26, "ALL"
            ),
            "UL_CE": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/UL_CE",
                51,
                None,
                ["UL", "CE"],
                None,
                None,
                "ALL",
            ),
            "NbPhases": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/NbPhases", 52, "ALL"
            ),
            "InputCurrent": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/InputSupply/InputCurrent", 53, "ALL"
            ),
            "InputMenu": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/InputSupply/InputMenu",
                74,
                None,
                ["Standard", "DualPack"],
                None,
                None,
                "ALL",
            ),
            "FiltFreq": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltFreq", 3, "ALL"
            ),
            "FiltStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltStart",
                4,
                "ALL",
            ),
            "FiltDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/FiltDur", 6, "ALL"
            ),
            "FiltDur2": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/FiltDur2",
                23,
                "ALL",
            ),
            "EconStart": GeckoTimeStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconStart",
                8,
                "ALL",
            ),
            "EconDur": GeckoTimeStructAccessor(
                self.struct, "ConfigStructure/SpaPackInternalOptions/EconDur", 10, "ALL"
            ),
            "EconProgAvailable": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconProgAvailable",
                71,
                None,
                ["NA", "STANDARD", "OUTSIDE_FILTER"],
                None,
                None,
                "ALL",
            ),
            "UDProgEcon": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/UDProgEcon",
                82,
                0,
                "ALL",
            ),
            "EconControlableManually": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/EconControlableManually",
                72,
                2,
                "ALL",
            ),
            "SoakOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SoakOnCustomKey",
                72,
                0,
                "ALL",
            ),
            "OffOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/OffOnCustomKey",
                72,
                1,
                "ALL",
            ),
            "CleanupOnCustomKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/CleanupOnCustomKey",
                72,
                3,
                "ALL",
            ),
            "MultiKeyOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/MultiKeyOption",
                76,
                None,
                ["NoBlowerOnI2C", "BlowerOnI2C"],
                None,
                None,
                "ALL",
            ),
            "MasterSlave": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/MasterSlave",
                73,
                None,
                ["", "Master", "Slave"],
                None,
                None,
                "ALL",
            ),
            "SlaveConfig": GeckoByteStructAccessor(
                self.struct,
                "ConfigStructure/SpaPackInternalOptions/SlaveConfig",
                75,
                "ALL",
            ),
            "TimeFormat": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/TimeFormat",
                34,
                None,
                ["NA", "AmPm", "24h"],
                None,
                None,
                "ALL",
            ),
            "AmbiantOHTrigADC": GeckoWordStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/AmbiantOHTrigADC", 64, "ALL"
            ),
            "Pump1UserAccess": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/Pump1UserAccess",
                81,
                1,
                ["BothSpeeds", "HighSpeedOnly"],
                None,
                2,
                "ALL",
            ),
            "SelfCleanMsg": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/SelfCleanMsg", 81, 4, "ALL"
            ),
            "BlowerKeyOption": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/BlowerKeyOption",
                81,
                5,
                ["FreePumpKey", "LastPumpKey"],
                None,
                2,
                "ALL",
            ),
            "KeypadTherapySupport": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/KeypadTherapySupport",
                43,
                0,
                "ALL",
            ),
            "CustomKeyEnabled": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/CustomKeyEnabled", 43, 1, "ALL"
            ),
            "ConfigChange": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/ConfigChange",
                43,
                2,
                ["NO_RESTRICTION", "PASSWORD_PROTECTED"],
                None,
                2,
                "ALL",
            ),
            "BreakerChange": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/BreakerChange",
                43,
                3,
                ["NO_RESTRICTION", "PASSWORD_PROTECTED"],
                None,
                2,
                "ALL",
            ),
            "KeypadBacklightColor": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/KeypadBacklightColor",
                43,
                4,
                ["OFF", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN", "WHITE"],
                None,
                8,
                "ALL",
            ),
            "KeypadBacklightEdit": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/KeypadBacklightEdit",
                43,
                7,
                ["Disable", "Enable"],
                None,
                2,
                "ALL",
            ),
            "CustomerID": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/CustomerID",
                44,
                None,
                [
                    "Generic",
                    "Hydropool",
                    "EndlessPools",
                    "Wellis",
                    "Alps",
                    "Artesian",
                    "Arctic",
                    "Barefoot",
                    "Beachcomber",
                    "Bellagio",
                    "Leisure_Prod_Ind",
                    "Bullfrog",
                    "Coast",
                    "Dimension_one",
                    "Dynasty",
                    "Four_Wind",
                    "Hotspring",
                    "Jacuzzi",
                    "Jazzi",
                    "LA",
                    "Pro_Float",
                    "MAAX",
                    "Marquis",
                    "PDC",
                    "Premium_Leisure",
                    "Strong",
                    "Sunrans",
                    "Sunrise",
                    "SuperiorSpas",
                    "Spa_Industries",
                    "Viking",
                    "WWO_Whirlcare",
                    "Okeanos",
                    "Clearwater",
                    "Deluxe",
                    "Idol",
                    "Aspen",
                    "ThermoSpas",
                    "Titan_Spas",
                    "IberSpa",
                    "Master_Spas",
                ],
                None,
                None,
                "ALL",
            ),
            "ModeKeyAsInvertDisplayKey": GeckoBoolStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/ModeKeyAsInvertDisplayKey",
                45,
                1,
                "ALL",
            ),
            "InfoMsgConfig": GeckoEnumStructAccessor(
                self.struct,
                "ConfigStructure/MiscCfg/InfoMsgConfig",
                45,
                2,
                ["HIDE_DETAILED_MSG", "SHOW_ALL_MSG", "", ""],
                None,
                4,
                "ALL",
            ),
            "LowerSetpointMenu": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/LowerSetpointMenu", 45, 4, "ALL"
            ),
            "KeypadOptions4": GeckoByteStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/KeypadOptions4", 46, "ALL"
            ),
            "DealerLockSupport": GeckoBoolStructAccessor(
                self.struct, "ConfigStructure/MiscCfg/DealerLockSupport", 47, 0, "ALL"
            ),
        }



"""Gecko Async Locator class."""

import asyncio
import logging
import socket
import time

from .async_spa_descriptor import GeckoAsyncSpaDescriptor
from .async_taskman import GeckoAsyncTaskMan
from .config import GeckoConfig, config_sleep
from .const import GeckoConstants
from .driver import (
    GeckoAsyncUdpProtocol,
    GeckoHelloProtocolHandler,
    Observable,
)
from .spa_events import GeckoSpaEvent

_LOGGER = logging.getLogger(__name__)


class GeckoAsyncLocator(Observable):
    """GeckoAsyncLocator class locates in.touch2 devices on your local LAN."""

    def __init__(
        self,
        taskman: GeckoAsyncTaskMan,
        event_handler: GeckoSpaEvent.CallBack,
        **kwargs: str | None,
    ) -> None:
        """Init the async locator."""
        super().__init__()
        # Get the arguments
        self._task_man: GeckoAsyncTaskMan = taskman
        self._event_handler: GeckoSpaEvent.CallBack = event_handler
        self._spa_address: str | None = kwargs.get("spa_address")
        if self._spa_address == "":
            self._spa_address = None
        self._spa_identifier: str | None = kwargs.get("spa_identifier")
        if self._spa_identifier == "":
            self._spa_identifier = None

        self._spas: list[GeckoAsyncSpaDescriptor] | None = None
        self._spa_identifiers: list[bytes] = []

        self._has_found_spa: bool = False
        self._started: float | None = None
        self._transport: asyncio.BaseTransport | None = None
        self._protocol: GeckoAsyncUdpProtocol | None = None

    async def _async_on_discovered(
        self, handler: GeckoHelloProtocolHandler, sender: tuple
    ) -> None:
        if handler.spa_identifier in self._spa_identifiers:
            return

        _LOGGER.debug("Discovered spa %s", handler.spa_identifier)
        if (
            self._spa_identifier is not None
            and self._spa_identifier
            != handler.spa_identifier.decode(GeckoConstants.MESSAGE_ENCODING)
        ):
            _LOGGER.debug("But we're not interested in that, so ignore it")
            return

        self._on_change(self)
        self._spa_identifiers.append(handler.spa_identifier)
        descriptor = GeckoAsyncSpaDescriptor(
            handler.spa_identifier,
            handler.spa_name,
            sender,
        )

        assert self._spas is not None  # noqa: S101
        self._spas.append(descriptor)
        await self._event_handler(
            GeckoSpaEvent.LOCATING_DISCOVERED_SPA, spa_descriptor=descriptor
        )

        if self._spa_address is not None or self._spa_identifier is not None:
            _LOGGER.debug(
                "Spa address or identifier was specified, so spa must have been found"
            )
            self._has_found_spa = True

    @property
    def age(self) -> float:
        """Get the ago of this locator."""
        if self._started is None:
            return 0
        return time.monotonic() - self._started

    @property
    def spas(self) -> list[GeckoAsyncSpaDescriptor] | None:
        """Get the list of spas."""
        return self._spas

    @property
    def has_had_enough_time(self) -> bool:
        """Return if we have had enough time to discover the spas."""
        return self.age > GeckoConfig.DISCOVERY_INITIAL_TIMEOUT_IN_SECONDS

    @property
    def is_running(self) -> bool:
        """Return the running state."""
        if self._started is None:
            return False
        return self._protocol is not None

    async def _broadcast_loop(self, hello_handler: GeckoHelloProtocolHandler) -> None:
        """Send a discovery message every second."""
        try:
            while True:
                if self._protocol is not None:
                    self._protocol.queue_send(
                        hello_handler,
                        GeckoHelloProtocolHandler.broadcast_address(
                            static_ip=self._spa_address
                        ),
                    )
                await config_sleep(
                    GeckoConstants.ASYNC_LOCATOR_BROADCAST_SLEEP,
                    "Async locator broadcast loop",
                )
        except asyncio.CancelledError:
            _LOGGER.debug("Broadcast loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("Broadcast loop caught exception")
            raise

    async def discover(self) -> None:
        """Discover spas on the local lan."""
        loop = asyncio.get_running_loop()
        on_con_lost = asyncio.Event()
        self._transport, _protocol = await loop.create_datagram_endpoint(
            lambda: GeckoAsyncUdpProtocol(self._task_man, on_con_lost, None),
            family=socket.AF_INET,
            allow_broadcast=True,
        )
        try:
            assert isinstance(_protocol, GeckoAsyncUdpProtocol)  # noqa: S101
            self._protocol = _protocol
            assert self._transport is not None  # noqa: S101
            self._spas = []

            hello_handler = GeckoHelloProtocolHandler.broadcast(
                async_on_handled=self._async_on_discovered
            )
            self._task_man.add_task(
                hello_handler.consume(self._protocol), "Hello handler", "LOC"
            )
            self._task_man.add_task(
                self._broadcast_loop(hello_handler), "Broadcast loop", "LOC"
            )

            self._started = time.monotonic()
            self._on_change(self)

            while self.age < GeckoConfig.DISCOVERY_TIMEOUT_IN_SECONDS:
                if self.has_had_enough_time and len(self._spas) > 0:
                    _LOGGER.info("Found %d spas ... %s", len(self._spas), self._spas)
                    break
                if self._has_found_spa:
                    break
                await config_sleep(
                    GeckoConstants.ASYNC_LOCATOR_BROADCAST_SLEEP,
                    "Async locator discovery loop",
                )

        finally:
            _LOGGER.debug("Discovery complete, close transport")
            self._task_man.cancel_key_tasks("LOC")
            self._transport.close()
            self._transport = None
            self._protocol = None
            self._on_change(self)



"""GeckoLib automation interface."""

from .async_facade import GeckoAsyncFacade
from .bainultra import BainUltra
from .base import GeckoAutomationBase, GeckoAutomationFacadeBase
from .blower import GeckoBlower
from .bubblegen import GeckoBubbleGenerator
from .button import GeckoButton
from .heater import GeckoWaterHeater, GeckoWaterHeaterAbstract
from .heatpump import GeckoHeatPump
from .ingrid import GeckoInGrid
from .inmix import GeckoInMix, GeckoInMixSynchro, GeckoInMixZone
from .keypad import GeckoKeypad
from .keypad_backlight import GeckoKeypadBacklight
from .light import GeckoLight
from .mrsteam import MrSteam
from .number import GeckoNumber
from .pump import GeckoPump
from .reminders import GeckoReminders
from .sensors import GeckoBinarySensor, GeckoErrorSensor, GeckoSensor
from .switch import GeckoSwitch
from .watercare import GeckoWaterCare
from .waterfall import GeckoWaterfall

__all__ = [
    "BainUltra",
    "GeckoAsyncFacade",
    "GeckoAutomationBase",
    "GeckoAutomationFacadeBase",
    "GeckoBinarySensor",
    "GeckoBlower",
    "GeckoBubbleGenerator",
    "GeckoButton",
    "GeckoErrorSensor",
    "GeckoHeatPump",
    "GeckoInGrid",
    "GeckoInMix",
    "GeckoInMixSynchro",
    "GeckoInMixZone",
    "GeckoKeypad",
    "GeckoKeypadBacklight",
    "GeckoLight",
    "GeckoNumber",
    "GeckoPump",
    "GeckoReminders",
    "GeckoSensor",
    "GeckoSwitch",
    "GeckoWaterCare",
    "GeckoWaterHeater",
    "GeckoWaterHeaterAbstract",
    "GeckoWaterfall",
    "MrSteam",
]



"""Automation heatpump class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.const import GeckoConstants

from .select import GeckoSelect

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade
    from geckolib.driver.accessor import (
        GeckoBoolStructAccessor,
    )

_LOGGER = logging.getLogger(__name__)


class GeckoHeatPump(GeckoSelect):
    """The heatpump mode."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize the heatpump class."""
        super().__init__(facade, "Heat Pump", GeckoConstants.KEY_COOLZONE_MODE)

        if GeckoConstants.KEY_MODBUS_HEATPUMP_DETECTED not in self._spa.accessors:
            self.set_availability(is_available=False)
        else:
            modbus_heatpump_detected: GeckoBoolStructAccessor = self._spa.accessors[
                GeckoConstants.KEY_MODBUS_HEATPUMP_DETECTED
            ]
            if not modbus_heatpump_detected.value:
                self.set_availability(is_available=False)

        # Set of mappings of constants to UI options
        self.set_mapping(
            {
                "CHILL": "Cool",
                "INTERNAL_HEAT": "Electric",
                "HEAT_SAVER": "Eco Heat",
                "HEAT_W_BOOST": "Smart Heat",
                "AUTO_SAVER": "Eco Auto",
                "AUTO_W_BOOST": "Smart Auto",
            }
        )

    @property
    def state(self) -> str:
        """Get the current state via the mapping."""
        if self._accessor is None:
            return "(Unknown)"
        return self.mapping[self._accessor.value]

    async def async_set_state(self, new_state: str) -> None:
        """Set the state of the select entity."""
        if self._accessor is None:
            _LOGGER.warning("%s can't set state with no accessor", self)
            return
        if new_state in self.reverse:
            new_state = self.reverse[new_state]
        await self._accessor.async_set_value(new_state)

    @property
    def states(self) -> list[str]:
        """Get the possible states."""
        if self._accessor is None:
            return []
        return list(self.mapping.values())



"""Observable class."""

import logging
from collections.abc import Callable
from typing import Any

_LOGGER = logging.getLogger(__name__)


class Observable:
    """
    Class to manage observables.

    Any class derived from this will support the ability to watch for changes.
    """

    def __init__(self) -> None:
        """Initialize the observable."""
        self._observers: list[Callable[[Any, Any, Any], None]] = []

    def watch(self, observer: Callable[[Any, Any, Any], None]) -> None:
        """Add an observer to this observable class."""
        if observer in self._observers:
            _LOGGER.warning(
                "Observer %s already in list, not going to add again", observer
            )
            return
        self._observers.append(observer)

    def unwatch(self, observer: Callable[[Any, Any, Any], None]) -> None:
        """Remove an observer to this observable class."""
        self._observers.remove(observer)

    def unwatch_all(self) -> None:
        """Remove all observers on this observable class."""
        _LOGGER.debug("Remove all observers from %r", self)
        self._observers.clear()

    def _on_change(
        self, sender: Any = None, old_value: Any = None, new_value: Any = None
    ) -> None:
        """Trigger the change notification for all observers."""
        _LOGGER.debug(
            f"{self.__class__.__name__} {sender} changed "  # noqa: G004
            f"from {old_value} to {new_value}"
        )
        for observer in self._observers:
            observer(sender, old_value, new_value)

    @property
    def has_observers(self) -> bool:
        """Determine if this has observers."""
        return len(self._observers) > 0

    def __repr__(self) -> str:
        """Get string reprosentation."""
        return f"{self.__class__.__name__} watched by={self._observers!r}"



"""Unit tests for the UDP protocol handlers."""  # noqa: INP001

# ruff: noqa: E501

import unittest
import unittest.mock

from context import (
    GeckoConfigFileProtocolHandler,
)

PARMS = (1, 2, b"SRCID", b"DESTID")


class TestGeckoConfigFileHandler(unittest.TestCase):
    """Config file tests."""

    def test_send_construct_request(self) -> None:
        handler = GeckoConfigFileProtocolHandler.request(1, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>SFILE\x01</DATAS></PACKT>",
        )

    def test_send_construct_response(self) -> None:
        handler = GeckoConfigFileProtocolHandler.response("inXM", 7, 8, parms=PARMS)
        self.assertEqual(
            handler.send_bytes,
            b"<PACKT><SRCCN>DESTID</SRCCN><DESCN>SRCID</DESCN>"
            b"<DATAS>FILES,inXM_C07.xml,inXM_S08.xml</DATAS></PACKT>",
        )

    def test_recv_can_handle(self) -> None:
        handler = GeckoConfigFileProtocolHandler()
        self.assertTrue(handler.can_handle(b"SFILE", PARMS))
        self.assertTrue(handler.can_handle(b"FILES", PARMS))
        self.assertFalse(handler.can_handle(b"OTHER", PARMS))

    def test_recv_handle_request(self) -> None:
        handler = GeckoConfigFileProtocolHandler()
        handler.handle(b"SFILE\x01", PARMS)
        self.assertFalse(handler.should_remove_handler)
        self.assertEqual(handler._sequence, 1)

    def test_recv_handle_response(self) -> None:
        handler = GeckoConfigFileProtocolHandler()
        handler.handle(b"FILES,inXM_C09.xml,inXM_S09.xml", PARMS)
        self.assertTrue(handler.should_remove_handler)
        self.assertEqual(handler.plateform_key, "inXM")
        self.assertEqual(handler.config_version, 9)
        self.assertEqual(handler.log_version, 9)

    @unittest.expectedFailure
    def test_recv_handle_response_error(self) -> None:
        handler = GeckoConfigFileProtocolHandler()
        self.assertTrue(handler.handle(b"FILES,inXM_C09.xml,inYE_S09.xml", PARMS))


if __name__ == "__main__":
    unittest.main()



"""GeckoSimulator class."""

# ruff: noqa: T201

import asyncio
import logging
import random
import socket
import time
from pathlib import Path
from typing import Any, Self

from geckolib import VERSION
from geckolib.async_taskman import GeckoAsyncTaskMan
from geckolib.config import config_sleep, release_config_change_waiters
from geckolib.const import GeckoConstants
from geckolib.driver import (
    GeckoAsyncPartialStatusBlockProtocolHandler,
    GeckoAsyncStructure,
    GeckoAsyncUdpProtocol,
    GeckoConfigFileProtocolHandler,
    GeckoGetChannelProtocolHandler,
    GeckoGetWatercareModeProtocolHandler,
    GeckoHelloProtocolHandler,
    GeckoPackCommandProtocolHandler,
    GeckoPacketProtocolHandler,
    GeckoPingProtocolHandler,
    GeckoRemindersProtocolHandler,
    GeckoReminderType,
    GeckoRFErrProtocolHandler,
    GeckoSetWatercareModeProtocolHandler,
    GeckoStatusBlockProtocolHandler,
    GeckoStructAccessor,
    GeckoUdpProtocolHandler,
    GeckoUnhandledProtocolHandler,
    GeckoUpdateFirmwareProtocolHandler,
    GeckoVersionProtocolHandler,
    GeckoWatercareScheduleManager,
)
from geckolib.driver.protocol.watercare import (
    GeckoGetWatercareScheduleListProtocolHandler,
)

from .shared_command import GeckoCmd
from .simulator_action import GeckoSimulatorAction
from .snapshot import GeckoSnapshot

_LOGGER = logging.getLogger(__name__)


SPA_IDENTIFIER = b"SPA01:02:03:04:05:06"


class GeckoSimulator(GeckoCmd, GeckoAsyncTaskMan):
    """
    GeckoSimulator.

    This is a server application to aid with investigating
    the Gecko protocol.
    """

    _STATUS_BLOCK_SEGMENT_SIZE = 39

    def __init__(self) -> None:
        """Initialize the simulator class."""
        GeckoAsyncTaskMan.__init__(self)
        GeckoCmd.__init__(self, self)

        # Async properties
        self._con_lost = asyncio.Event()
        self._protocol: GeckoAsyncUdpProtocol | None = None
        self._transport: asyncio.BaseTransport | None = None
        self._name: str = ""
        self.structure: GeckoAsyncStructure = GeckoAsyncStructure(
            self._async_on_set_value
        )
        self.snapshot: GeckoSnapshot = GeckoSnapshot()
        self._reliability = 1.0
        self._tardiness = 0.0
        self._do_rferr = False
        self._do_thermodynamics = False
        self._send_structure_change = False
        self._clients = []
        random.seed()

        self._accessor_change_queue: asyncio.Queue = asyncio.Queue()

        self._structure_change_queue: asyncio.Queue = asyncio.Queue()
        self._can_report_structure_changes: asyncio.Event = asyncio.Event()
        self._can_report_structure_changes.set()

        self.intro = (
            "Welcome to the Gecko simulator. Type help or ? to list commands.\n"
        )
        self.prompt = "(GeckoSim) "
        try:
            import readline

            readline.set_completer_delims(" \r\n")
        except ImportError:
            pass

        self._action = GeckoSimulatorAction()
        for name in GeckoSimulatorAction.__annotations__:
            setattr(self._action, name, None)

        self._current_watercare_mode: int = GeckoConstants.WATERCARE_MODE[1]
        self._reminders: list[tuple[GeckoReminderType, int]] = [
            (GeckoReminderType.RINSE_FILTER, -13),
            (GeckoReminderType.CLEAN_FILTER, 0),
            (GeckoReminderType.CHANGE_WATER, 47),
            (GeckoReminderType.CHECK_SPA, 687),
            (GeckoReminderType.INVALID, -13),
            (GeckoReminderType.INVALID, -13),
            (GeckoReminderType.INVALID, 0),
            (GeckoReminderType.INVALID, 0),
            (GeckoReminderType.INVALID, 0),
            (GeckoReminderType.INVALID, 0),
        ]

        self._watercare_schedule_manager: GeckoWatercareScheduleManager = (
            GeckoWatercareScheduleManager()
        )

        self._hello_task: asyncio.Task | None = None

    async def __aenter__(self) -> Self:
        """Support async with."""
        await GeckoAsyncTaskMan.__aenter__(self)
        GeckoAsyncTaskMan.add_task(
            self, self._process_accesor_changes(), "Accessor Change Handler", "SIM"
        )
        GeckoAsyncTaskMan.add_task(
            self, self._notify_structure_changes(), "Notify Structure Changes", "SIM"
        )
        self.add_task(self._timer_loop(), "Timer", "SIM")
        return self

    async def __aexit__(self, *_args: object) -> None:
        """Support 'with' statements."""
        await self.do_stop("")

    def do_about(self, _arg: str) -> None:
        """Display information about this client program and support library : about."""
        print()
        print(
            "GeckoSimulator: A python program using GeckoLib library to simulate Gecko"
            " enabled devices with in.touch2 communication modules"
        )
        print(f"Library version v{VERSION}")

    async def do_start(self, _args: str) -> None:
        """Start the configured simulator : start."""
        loop = asyncio.get_running_loop()
        self._con_lost.clear()
        # Create a socket that can handle broadcast
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        sock.bind(("0.0.0.0", 10022))  # noqa: S104
        # Start the transport and protocol handler
        self._transport, self._protocol = await loop.create_datagram_endpoint(
            lambda: GeckoAsyncUdpProtocol(self, self._con_lost, None),
            sock=sock,
        )
        await self._install_standard_handlers()
        self._send_structure_change = True

    async def do_stop(self, _args: str) -> None:
        """Stop the simulator : stop."""
        if self._protocol:
            self._protocol.disconnect()
            self._protocol = None
        if self._transport:
            self._transport.close()
            self._transport = None

    def do_parse(self, args: str) -> None:
        """
        Parse logfiles.

        Extract snapshots to the ./snapshot directory. Will
        overwrite identically named snapshot files if present.

        usage: parse <logfile>
        """
        for snapshot in GeckoSnapshot.parse_log_file(args):
            snapshot.save("snapshots")
            print(f"Saved snapshot snapshots/{snapshot.filename}")

    def do_reliability(self, args: str) -> None:
        """
        Set simulator reliability factor.

        Reliability is a measure of how likely
        the simulator will respond to an incoming message. Reliability of 1.0 (default)
        means the simulator will always respond, whereas 0.0 means it will never
        respond. This does not take into account messages that actually don't get
        recieved : reliability <factor> where <factor> is a float between 0.0 and
        1.0.
        """
        if args == "":
            print(f"Current reliability is {self._reliability}")
            return
        self._reliability = min(1.0, max(0.0, float(args)))

    def do_tardiness(self, args: str) -> None:
        """
        Set simulator tardiness factor.

        Tardomess is a measure of how rapidly the simulator will respond to an
        incoming message. Tardiness of 0.0 (default) means the simulator will
        respond immediately. Larger numbers are how long in seconds it could
        take to respond. This isn't an absolute delay, it will be somewhere in
        the range 0.0 to the tardiness value.

        Usage: tardiness [<factor>] where <factor> is a float. Missing the
               parameter will show the current value
        """
        if args == "":
            print(f"Current tardiness is {self._tardiness}")
            return
        self._tardiness = float(args)

    def do_rferr(self, args: str) -> None:
        """Set the simulator to response with RFERR if the parameter is True."""
        self._do_rferr = args.lower() == "true"
        print(f"RFERR mode set to {self._do_rferr}")

    def complete_parse(
        self, text: str, _line: str, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the parse command."""
        return self._complete_path(text)

    async def do_load(self, args: str) -> None:
        """
        Load a snapshot.

        Usage: load <snapshot>
        """
        snapshots = GeckoSnapshot.parse_log_file(args)
        if len(snapshots) == 1:
            await self.set_snapshot(snapshots[0])
            return
        print(
            f"{args} contains {len(snapshots)} snapshots. Please use the"
            f" `parse` command to break it apart"
        )

    def complete_load(
        self, text: str, _line: str, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the load command."""
        return self._complete_path(text)

    def do_save(self, args: str) -> None:
        """
        Save a snapshot.

        Usage: save <snapshot file>
        """
        path = Path(args)
        data = self.get_snapshot_data()
        with path.open("w") as f:
            f.write(f"{data}")

    def complete_save(
        self, text: str, _line: str, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the save command."""
        return self._complete_path(text)

    async def do_import(self, args: str) -> None:
        """Import a JSON snapshot."""
        snapshot = GeckoSnapshot.parse_json(args)
        await self.set_snapshot(snapshot)

    async def do_name(self, args: str) -> None:
        """Set the name of the spa : name <spaname>."""
        if args == "":
            print(f"Spa name is {self._name}")
            return

        self._name = args
        if self._hello_task is not None:
            self._hello_task.cancel()
            self._hello_task = None
            release_config_change_waiters()
            await asyncio.sleep(1)

        if self._protocol is not None:
            # Hello handler
            self._hello_task = self.add_task(
                GeckoHelloProtocolHandler.response(
                    SPA_IDENTIFIER,
                    self._name,
                    async_on_handled=self._async_on_hello,
                ).consume(self._protocol),
                "Hello handler",
                "SIM",
            )

    async def set_snapshot(self, snapshot: GeckoSnapshot) -> None:
        """Assign snapshot to this simulator."""
        compatible = self.snapshot.is_compatible(snapshot)
        if not compatible:
            print("The new snapshot structure is not compatible, so doing a full reset")
            print(self.prompt, end="", flush=True)
            if "PackType" in self.structure.accessors:
                await self.structure.accessors["PackType"].async_set_value(
                    self.structure.accessors["PackType"].value, always_send=True
                )

            self.structure.reset()

        self.snapshot = snapshot
        if self.snapshot.spa_name is not None:
            await self.do_name(self.snapshot.spa_name)
        _LOGGER.debug(snapshot)

        await self._set_structure_from_snapshot(
            self.structure, self.snapshot, do_rebuild=not compatible
        )
        if not compatible:
            for accessor in self.structure.accessors.values():
                accessor.set_read_write("ALL")
                accessor.watch(self._on_accessor_changed)

    async def _set_structure_from_snapshot(
        self, struct: GeckoAsyncStructure, snapshot: GeckoSnapshot, *, do_rebuild: bool
    ) -> None:
        struct.replace_status_block_segment(0, snapshot.bytes)

        if do_rebuild:
            # Must rebuild stuff
            try:
                # Attempt to get config and log classes
                await struct.load_pack_class(snapshot.packtype.lower())
                await struct.load_config_module(snapshot.config_version)
                await struct.load_log_module(snapshot.log_version)
                await struct.check_for_accessories()
                struct.build_accessors()

            except ModuleNotFoundError:
                _LOGGER.exception("Module not found during snapshot load")
        struct.build_connections()

    async def do_diff(self, arg: str) -> None:
        """
        Perform a diff between spa data structures.

        You can either diff two separate files,
        or a single file and the current structure.

        usage: diff <file1> <file2>
        """
        file1 = arg
        file2 = ""
        rest = None
        if " " in arg:
            file1, file2, *rest = arg.split(" ")
        if not file1:
            print("Need at least one file.")

        struct = GeckoAsyncStructure(None)
        snapshot = GeckoSnapshot.parse_log_file(file1)[0]
        await self._set_structure_from_snapshot(struct, snapshot, do_rebuild=True)

        differences = self.structure.get_differences(struct)
        if differences:
            print("Differences are:")
            print("\n".join(differences))
        else:
            print("Identical snapshot")
        print(self.prompt, end="", flush=True)

    def complete_diff(
        self, text: str, _line: str, _start_idx: int, _end_idx: int
    ) -> list[str]:
        """Complete the do_diff command."""
        return self._complete_path(text)

    async def do_reset_to(self, arg: str) -> None:
        """
        Reset simulator to specific platform and version.

        usage: reset_to <platform> <config_version> <log_version>.
        """
        try:
            plateform, config, log = arg.split(" ")
            await self.set_snapshot(GeckoSnapshot.create(plateform, config, log))
        except IndexError:
            pass

    def do_thermodynamics(self, arg: str) -> None:
        """
        Run thermodynamics simulator.

        Turn the thermodynamics simulation control on or off
        usage: thermodynamics <ON|OFF>. Defaults to ON
        """
        if not arg:
            arg = "ON"
        self._do_thermodynamics = arg == "ON"

    async def _should_ignore(
        self,
        handler: GeckoUdpProtocolHandler,
        sender: tuple,
        *,
        respect_rferr: bool = True,
    ) -> bool:
        if respect_rferr and self._do_rferr:
            assert self._protocol is not None  # noqa: S101
            self._protocol.queue_send(
                GeckoRFErrProtocolHandler.response(parms=sender), sender
            )
            # Always ignore responses because we've already replied with RFERR
            return True

        tardiness = random.random() * self._tardiness  # noqa: S311 (Not interested in crypto strength!)
        if tardiness > 0.0:
            await asyncio.sleep(tardiness)

        should_ignore = random.random() > self._reliability  # noqa: S311 (Not interested in crypto strength!)
        if should_ignore:
            _LOGGER.debug(
                f"Unreliable simulator ignoring request for {handler} from {sender}"  # noqa: G004
            )
        return should_ignore

    async def _install_standard_handlers(self) -> None:
        """
        Install standard handlers.

        All simulators needs to have some basic functionality such
        as discovery, error handling et al
        """
        await self.do_name("Simulator" if self._name == "" else self._name)

        assert self._protocol is not None  # noqa: S101
        # Helper to unwrap PACK packets
        self.add_task(
            GeckoPacketProtocolHandler(async_on_handled=self._async_on_packet).consume(
                self._protocol
            ),
            "Packet handler",
            "SIM",
        )
        # Unhandled packets get thrown
        self.add_task(
            GeckoUnhandledProtocolHandler().consume(self._protocol),
            "Unhandled packet",
            "SIM",
        )
        # Ping response handler
        self.add_task(
            GeckoPingProtocolHandler(async_on_handled=self._async_on_ping).consume(
                self._protocol
            ),
            "Ping handler",
            "SIM",
        )
        # Version handler
        self.add_task(
            GeckoVersionProtocolHandler(
                async_on_handled=self._async_on_version
            ).consume(self._protocol),
            "Version handler",
            "SIM",
        )
        # Channel handler
        self.add_task(
            GeckoGetChannelProtocolHandler(
                async_on_handled=self._async_on_get_channel
            ).consume(self._protocol),
            "Get channel",
            "SIM",
        )
        # Config file
        self.add_task(
            GeckoConfigFileProtocolHandler(
                async_on_handled=self._async_on_config_file
            ).consume(self._protocol),
            "Config file",
            "SIM",
        )
        # Status block
        self.add_task(
            GeckoStatusBlockProtocolHandler(
                async_on_handled=self._async_on_status_block
            ).consume(self._protocol),
            "Status block",
            "SIM",
        )
        # Watercare if regular spa pack
        if self.structure.is_spa_pack:
            self.add_task(
                GeckoGetWatercareModeProtocolHandler(
                    async_on_handled=self._async_on_get_watercare_mode
                ).consume(self._protocol),
                "GetWatercareMode",
                "SIM",
            )
            self.add_task(
                GeckoSetWatercareModeProtocolHandler(
                    async_on_handled=self._async_on_set_watercare_mode
                ).consume(self._protocol),
                "SetWatercareMode",
                "SIM",
            )
            self.add_task(
                GeckoGetWatercareScheduleListProtocolHandler(
                    async_on_handled=self._async_on_get_watercare_schedule
                ).consume(self._protocol),
                "GetWatercareSchedule",
                "SIM",
            )

        # Reminders if regular spa pack
        if self.structure.is_spa_pack:
            # Reminders
            self.add_task(
                GeckoRemindersProtocolHandler(
                    async_on_handled=self._async_on_reminders
                ).consume(self._protocol),
                "Reminders",
                "SIM",
            )

        # Update firmware fake
        self.add_task(
            GeckoUpdateFirmwareProtocolHandler(
                async_on_handled=self._async_on_update_firmware
            ).consume(self._protocol),
            "Update Firmware",
            "SIM",
        )
        # Pack command
        self.add_task(
            GeckoPackCommandProtocolHandler(
                async_on_handled=self._async_on_pack_command
            ).consume(self._protocol),
            "Pack command",
            "SIM",
        )
        # Partial pack ack
        self.add_task(
            GeckoAsyncPartialStatusBlockProtocolHandler(self._protocol).consume(
                self._protocol
            ),
            "Partial pack ACK",
            "SIM",
        )

    async def _async_on_hello(
        self, handler: GeckoHelloProtocolHandler, sender: tuple
    ) -> None:
        if handler.was_broadcast_discovery:
            if await self._should_ignore(handler, sender, respect_rferr=False):
                return
            assert self._protocol is not None  # noqa: S101
            if (
                time.monotonic() - handler.last_response
                > GeckoConstants.SIMULATOR_MIN_TIME_BETWEEN_HELLO_BROADCAST_RESPONSES
            ):
                handler.last_response = time.monotonic()
                self._protocol.queue_send(handler, sender)
        elif handler.client_identifier is not None:
            # We're not fussy, we'll chat to anyone!
            pass

    async def _async_on_ping(
        self, handler: GeckoPingProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoPingProtocolHandler.response(on_get_parms=lambda _handler: sender),
            sender,
        )
        if sender not in self._clients:
            self._clients.append(sender)

    async def _async_on_packet(
        self, handler: GeckoPacketProtocolHandler, _sender: tuple
    ) -> None:
        assert self._protocol is not None  # noqa: S101
        assert handler.parms is not None  # noqa: S101
        if handler.parms[3] != SPA_IDENTIFIER:
            _LOGGER.warning(
                "Dropping packet from %s because it didn't match %s",
                handler.parms,
                SPA_IDENTIFIER,
            )
        assert handler.packet_content is not None  # noqa: S101
        self._protocol.datagram_received(handler.packet_content, handler.parms)

    async def _async_on_version(
        self, handler: GeckoVersionProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        assert self.snapshot is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoVersionProtocolHandler.response(
                self.snapshot.intouch_EN,
                self.snapshot.intouch_CO,
                parms=sender,
            ),
            sender,
        )

    async def _async_on_get_channel(
        self, handler: GeckoGetChannelProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoGetChannelProtocolHandler.response(10, 33, parms=sender),
            sender,
        )

    async def _async_on_config_file(
        self, handler: GeckoConfigFileProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        assert self.snapshot is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoConfigFileProtocolHandler.response(
                self.snapshot.packtype,
                self.snapshot.config_version,
                self.snapshot.log_version,
                parms=sender,
            ),
            sender,
        )

    async def _async_on_status_block(
        self, handler: GeckoStatusBlockProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        for idx, start in enumerate(
            range(
                handler.start,
                handler.start + handler.length,
                self._STATUS_BLOCK_SEGMENT_SIZE,
            )
        ):
            length = min(
                self._STATUS_BLOCK_SEGMENT_SIZE,
                len(self.structure.status_block) - start,
            )
            next_index = (idx + 1) % (
                (handler.length // self._STATUS_BLOCK_SEGMENT_SIZE) + 1
            )
            if await self._should_ignore(handler, sender, respect_rferr=False):
                continue
            assert self._protocol is not None  # noqa: S101
            self._protocol.queue_send(
                GeckoStatusBlockProtocolHandler.response(
                    idx,
                    next_index,
                    self.structure.status_block[start : start + length],
                    parms=sender,
                ),
                sender,
            )

    async def _async_on_get_watercare_mode(
        self, handler: GeckoGetWatercareModeProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoGetWatercareModeProtocolHandler.get_response(
                self._current_watercare_mode, parms=sender
            ),
            sender,
        )

    async def _async_on_set_watercare_mode(
        self, handler: GeckoSetWatercareModeProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        print(
            f"Set watercare mode to {GeckoConstants.WATERCARE_MODE_STRING[handler.mode]}"  # noqa: E501
        )
        self._current_watercare_mode = handler.mode
        self._protocol.queue_send(
            GeckoSetWatercareModeProtocolHandler.set_response(
                self._current_watercare_mode, parms=sender
            ),
            sender,
        )

    async def _async_on_get_watercare_schedule(
        self, handler: GeckoGetWatercareScheduleListProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoGetWatercareScheduleListProtocolHandler.get_response(
                self._watercare_schedule_manager, parms=sender
            ),
            sender,
        )

    async def _async_on_reminders(
        self, handler: GeckoRemindersProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        if handler.is_request:
            self._protocol.queue_send(
                GeckoRemindersProtocolHandler.req_response(
                    self._reminders,
                    parms=sender,
                ),
                sender,
            )
            return

        assert handler.reminders  # noqa: S101
        self._reminders = handler.reminders
        self._protocol.queue_send(
            GeckoRemindersProtocolHandler.set_ack(parms=sender), sender
        )

    async def _async_on_update_firmware(
        self, handler: GeckoUpdateFirmwareProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoUpdateFirmwareProtocolHandler.response(parms=sender), sender
        )

    async def _async_on_pack_command(
        self, handler: GeckoPackCommandProtocolHandler, sender: tuple
    ) -> None:
        if await self._should_ignore(handler, sender):
            return
        assert self._protocol is not None  # noqa: S101
        self._protocol.queue_send(
            GeckoPackCommandProtocolHandler.response(parms=sender), sender
        )
        if handler.is_key_press:
            await self._async_handle_pack_key_press(handler.keycode)
        elif handler.is_set_value:
            await self._async_handle_pack_set_value(
                handler.position, handler.length, handler.new_data
            )
        else:
            _LOGGER.error("Unhandled pack command")

    async def _async_handle_pack_key_press(self, keycode: int) -> None:
        """Handle a key press command."""
        _LOGGER.debug("Pack command press key %d", keycode)

        self._fill_action_state()

        # Iterate through all attributes in GeckoConstants
        for name, val in vars(GeckoConstants).items():
            # Check if the attribute is an integer, matches the given value,
            # and starts with 'KEYPAD_'
            if isinstance(val, int) and val == keycode and name.startswith("KEYPAD_"):
                # Now find a function in the action class
                func = getattr(self._action, f"on_{name}", None)
                if func is not None:
                    func()
                    await self._report_action_state()
                    return

        _LOGGER.warning("Simulator didn't handle key pad command for %d", keycode)

    async def _async_handle_pack_set_value(
        self, pos: int, _length: int, data: bytes
    ) -> None:
        """Handle set value."""
        _LOGGER.debug(
            "Replace status block segmemnt @ %d with %s (%d bytes)",
            pos,
            f"{data}",
            len(data),
        )
        self.structure.replace_status_block_segment(pos, data)
        if self._send_structure_change:
            await self._structure_change_queue.put(pos)

    async def _async_on_set_value(self, pos: int, length: int, newvalue: Any) -> None:
        """Call when the structure notifies a change."""
        change = (pos, GeckoStructAccessor.pack_data(length, newvalue))
        await self._async_handle_pack_set_value(change[0], length, change[1])

    async def _notify_structure_changes(self) -> None:
        try:
            while True:
                await self._can_report_structure_changes.wait()
                changes: list = []
                changes.append(await self._structure_change_queue.get())
                while self._structure_change_queue.qsize():
                    changes.append(self._structure_change_queue.get_nowait())
                # The change report system only handles two byte changes, so
                # we have to suck those from the actual status block.
                changes = [
                    (
                        change,
                        self.structure.status_block[change : change + 2],
                    )
                    for change in changes
                ]
                assert self._protocol is not None  # noqa: S101
                for client in self._clients:
                    self._protocol.queue_send(
                        GeckoAsyncPartialStatusBlockProtocolHandler.report_changes(
                            self._protocol, changes, parms=client
                        ),
                        client,
                    )

        except asyncio.CancelledError:
            _LOGGER.debug("Simulator notify structure change loop cancelled")
            raise

        except:
            _LOGGER.exception("Simulator notify structure change loop exception")
            raise

        finally:
            _LOGGER.debug("Simulator notify structure change loop finished")

    def _on_accessor_changed(
        self, accessor: GeckoStructAccessor, old_value: Any, new_value: Any
    ) -> None:
        print(f"accessor {accessor} changed from {old_value} to {new_value}")
        self._accessor_change_queue.put_nowait((accessor, old_value, new_value))
        if self._send_structure_change:
            self._structure_change_queue.put_nowait(accessor.pos)

    async def _process_accesor_changes(self) -> None:
        accessor: GeckoStructAccessor
        old_value: Any
        new_value: Any
        try:
            while True:
                accessor, old_value, new_value = await self._accessor_change_queue.get()

                self._fill_action_state()
                func = getattr(self._action, f"on_{accessor.tag}", None)
                if func is not None:
                    func()
                    await self._report_action_state()
                else:
                    _LOGGER.debug(
                        "%s changed from %s to %s", accessor, old_value, new_value
                    )
                self._accessor_change_queue.task_done()

        except asyncio.CancelledError:
            _LOGGER.debug("Simulator accessort change loop cancelled")
            raise

        except:
            _LOGGER.exception("Simulator value update loop exception")
            raise

        finally:
            _LOGGER.debug("Simulator value update loop finished")

    def _fill_action_state(self) -> None:
        for name in vars(self._action):
            if name in self.structure.accessors:
                setattr(self._action, name, self.structure.accessors[name].value)
            else:
                setattr(self._action, name, None)
        self._action.set_connections(self.structure.connections)

    async def _report_action_state(self) -> None:
        # Write all the properties back, only the changed
        # ones will be sent.
        self._can_report_structure_changes.clear()
        for name, val in vars(self._action).items():
            if name in self.structure.accessors and val is not None:
                await self.structure.accessors[name].async_set_value(val)
        self._can_report_structure_changes.set()

    def get_snapshot_data(self) -> dict:
        """Get snapshot data that the simulator can supply."""
        data = self.structure.get_snapshot_data()
        assert self.snapshot is not None  # noqa: S101
        data.update(
            {
                "intouch version EN": self.snapshot.intouch_EN_str,
                "intouch version CO": self.snapshot.intouch_CO_str,
            }
        )
        return data

    async def _timer_loop(self) -> None:
        elapsed_time = 0
        try:
            while True:
                await config_sleep(1, "Simulator Timer")
                elapsed_time += 1
                self._fill_action_state()
                self._action.every_second(elapsed_time)
                if elapsed_time % 10 == 0 and self._do_thermodynamics:
                    self._action.check_thermodynamics(elapsed_time)
                if elapsed_time % 60 == 0:
                    self._action.every_minute(int(elapsed_time / 60))
                if elapsed_time % 3600 == 0:
                    self._action.every_hour(int(elapsed_time / 3600))
                await self._report_action_state()

        except asyncio.CancelledError:
            _LOGGER.debug("Simulator timer loop cancelled")
            raise
        except Exception:
            _LOGGER.exception("Simulator timer loop caught exception")
            raise



"""Configuration class for complete async example."""

import configparser
import logging
from pathlib import Path

# Configuration file constants
CONFIG_FILE = "cui.ini"
CK_DEFAULT = "DEFAULT"
CK_SPA_ID = "SPA_ID"
CK_SPA_ADDR = "SPA_ADDR"
CK_SPA_NAME = "SPA_NAME"

_LOGGER = logging.getLogger(__name__)


class Config:
    """Hold the configuration for the complete async example app."""

    def __init__(self) -> None:
        """Initialize the config class."""
        _LOGGER.debug("Read config from %s", CONFIG_FILE)
        self._config = configparser.ConfigParser()
        self._config.read(CONFIG_FILE)

    def save(self) -> None:
        """Save the config file."""
        with Path(CONFIG_FILE).open("w") as configfile:
            self._config.write(configfile)

    @property
    def spa_id(self) -> str | None:
        """Get the spa id."""
        return self._config[CK_DEFAULT].get(CK_SPA_ID, None)

    def set_spa_id(self, spa_id: str | None) -> None:
        """Set the spa id."""
        if spa_id is None:
            self._config.remove_option(CK_DEFAULT, CK_SPA_ID)
        else:
            self._config[CK_DEFAULT][CK_SPA_ID] = spa_id

    @property
    def spa_address(self) -> str | None:
        """Get the spa address."""
        return self._config[CK_DEFAULT].get(CK_SPA_ADDR, None)

    def set_spa_address(self, spa_address: str | None) -> None:
        """Set the spa address."""
        if spa_address is None:
            self._config.remove_option(CK_DEFAULT, CK_SPA_ADDR)
        else:
            self._config[CK_DEFAULT][CK_SPA_ADDR] = spa_address

    @property
    def spa_name(self) -> str | None:
        """Get the spa name."""
        return self._config[CK_DEFAULT].get(CK_SPA_NAME, None)

    def set_spa_name(self, spa_name: str | None) -> None:
        """Set the spa name."""
        if spa_name is None:
            self._config.remove_option(CK_DEFAULT, CK_SPA_NAME)
        else:
            self._config[CK_DEFAULT][CK_SPA_NAME] = spa_name



"""Gecko Watercare."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.const import GeckoConstants

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoWaterCare(GeckoAutomationFacadeBase):
    """Watercare manangement class."""

    def __init__(self, facade: GeckoAsyncFacade) -> None:
        """Initialize watercare class."""
        super().__init__(facade, "WaterCare", "WATERCARE")
        self.active_mode: int | None = None
        if facade.spa.struct.is_spa_pack:
            self.set_availability(is_available=True)

    @property
    def mode(self) -> int | None:
        """Return the active water care mode."""
        return self.active_mode

    @property
    def modes(self) -> list[str]:
        """Return all the possible water care modes."""
        return GeckoConstants.WATERCARE_MODE_STRING

    @property
    def state(self) -> str:
        """Return the current state."""
        if self.mode is None:
            return "Unknown"
        return GeckoConstants.WATERCARE_MODE_STRING[self.mode]

    @property
    def states(self) -> list[str]:
        """Return all the states."""
        return self.modes

    async def async_set_state(self, new_mode: str | int) -> None:
        """Set the state. Support select style objects."""
        await self.async_set_mode(new_mode)

    async def async_set_mode(self, new_mode: str | int) -> None:
        """
        Set the active watercare mode to new_mode.

        new_mode can be a string, in which case the value must be a member of
        GeckoConstants.WATERCARE_MODE_STRING, or it can be an integer from
        GeckoConstants.WATERCARE_MODE
        """
        if isinstance(new_mode, str):
            new_mode = GeckoConstants.WATERCARE_MODE_STRING.index(new_mode)
        await self._spa.async_set_watercare_mode(new_mode)
        self.change_watercare_mode(new_mode)

    def change_watercare_mode(self, new_mode: int) -> None:
        """Change the watercare mode."""
        if self.active_mode != new_mode:
            old_mode = self.active_mode
            self.active_mode = new_mode
            self._on_change(self, old_mode, self.active_mode)

    def __str__(self) -> str:
        """Stringise the class."""
        if not self.is_available:
            return f"{self.name}: Unavailable"
        if self.active_mode is None:
            return f"{self.name}: Waiting..."
        if self.active_mode < 0 or self.active_mode > len(
            GeckoConstants.WATERCARE_MODE_STRING
        ):
            return f"Unknown Water care mode (index:{self.active_mode})"
        return f"{self.name}: {GeckoConstants.WATERCARE_MODE_STRING[self.active_mode]}"

    @property
    def monitor(self) -> str:
        """Get monitor string."""
        if not self.is_available:
            return "WC: None"
        if self.active_mode is None:
            return "WC: ?"
        return f"WC: {self.active_mode}"



"""Automation number class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from geckolib.driver.accessor import GeckoStructAccessor

from .base import GeckoAutomationFacadeBase

if TYPE_CHECKING:
    from geckolib.automation.async_facade import GeckoAsyncFacade

_LOGGER = logging.getLogger(__name__)


class GeckoNumber(GeckoAutomationFacadeBase):
    """A number object can input a value in a range and can report the current state."""

    def __init__(
        self,
        facade: GeckoAsyncFacade,
        name: str,
        tag: str,
        unit_accessor: GeckoStructAccessor | str | None = None,
    ) -> None:
        """Initialize the number class."""
        super().__init__(facade, name, tag)
        self._accessor = None
        self.native_min_value: float = 0.0
        self.native_max_value: float = 100.0
        self.native_step: float = 1.0
        self.mode = "auto"
        if tag in facade.spa.accessors:
            self._accessor = facade.spa.accessors[tag]
            self._accessor.watch(self._on_change)
            self.set_availability(is_available=True)
        self._unit_accessor = unit_accessor
        if isinstance(self._unit_accessor, GeckoStructAccessor):
            self._unit_accessor.watch(self._on_change)

    @property
    def native_value(self) -> float:
        """Get the current value."""
        if self._accessor is None:
            return 0.0
        return self._accessor.value

    async def async_set_native_value(self, new_value: float) -> None:
        """Set the value of the number entity."""
        if self._accessor is None:
            _LOGGER.warning("%s can't set value with no accessor.", self)
            return
        _LOGGER.debug("%s async set value %f", self.name, new_value)
        await self._accessor.async_set_value(int(new_value))

    @property
    def native_unit_of_measurement(self) -> str | None:
        """The unit of measurement for the sensor, or None."""
        if isinstance(self._unit_accessor, GeckoStructAccessor):
            return self._unit_accessor.value
        return self._unit_accessor

    def __str__(self) -> str:
        """Stringize the class."""
        return f"{self.name}: {self.native_value}{self.native_unit_of_measurement}"

    @property
    def monitor(self) -> str:
        """Get monitore string."""
        return f"{self.key}: {self.native_value}{self.native_unit_of_measurement}"
